import numpy as np


def uij(wf1,wf2):
  """ Calcultes the matrix product of two sets of input wavefunctions"""
  out =  np.matrix(np.conjugate(wf1))@(np.matrix(wf2).T)  # faster way
  return out


def uij_slow(wf1,wf2):
  m = np.matrix(np.zeros((len(wf1),len(wf2)),dtype=np.complex_))
  for i in range(len(wf1)):
    for j in range(len(wf2)):
      m[i,j] = np.conjugate(wf1[i]).dot(wf2[j])
  return m





import numpy as np
from numba import jit



def get_mf_anomalous(v,dm):
    """Compute the anomalous part of the mean-field,
    input in the anomalous density matrix in Nambu basis"""
    zero = dm[(0,0,0)]*0. # zero
    mf = dict()
    for d in v: mf[d] = zero.copy()  # initialize
    for d in v: # loop over directions
        d2 = tuple(-np.array(d))
        m = anomalous_term_ij(2*v[d],dm[d2]) # get matrix
        mf[d] = mf[d] + m # add normal term
    return mf



def anomalous_term_ij(v,dm):
    """Return the anomalous term of the mean-field,
    assuming a nmabu basis"""
    # we will assume that v contains up,down in alternating order
    n = dm.shape[0] # number of spinless sites
    out = np.zeros((n,n),dtype=np.complex_)
    return anomalous_term_ij_jit(v,dm,out)



# this anomalous term enforces even superconductivity

@jit(nopython=True)
def anomalous_term_ij_jit(v,dm,out):
    ns = len(dm)//2 # number of spinless sites
    for i in range(ns): # loop over spinless sites
        for j in range(ns): # loop over spinless sites
          out[2*i,2*j] = v[2*i,2*j+1]*dm[2*j,2*i]  # down,up
          out[2*i,2*j+1] = v[2*i,2*j]*dm[2*j+1,2*i]  # up,up
          out[2*i+1,2*j+1] = v[2*i+1,2*j]*dm[2*j+1,2*i+1]  # up,down
          out[2*i+1,2*j] = v[2*i+1,2*j+1]*dm[2*j,2*i+1]  # down,down
    return out



def enforce_eh_symmetry_anomalous(d01):
    """Enforce electron-hole symmetry in the two sectors"""
    d01 = enforce_eh_symmetry_anomalous_sector(d01)
    d10 = enforce_eh_from_sector(d01)
    return d01,d10



def enforce_eh_from_sector(d):
    """Given one sector of the Hamiltonian, return the other one"""
    out = dict() # dictionary
    for key in d:
        m = d[key] # one key
        o01 = m*0.0
        key2 = tuple(-np.array(key)) # the opposite
        out[key2] = enforce_eh_from_sector_jit(m,o01)
    return out # return


@jit(nopython=True)
def enforce_eh_from_sector_jit(d,o):
    """Given the ee sector, return the hh sector"""
    return np.conjugate(d.T) # hermitian conjugate
#    ns = len(d)//2 # number of sites
#    for i in range(ns): # loop
#        for j in range(ns): # loop
#            o[2*i+1,2*j+1] = np.conjugate(d[2*j+1,2*i+1])      # ud
#            o[2*i,2*j] = np.conjugate(d[2*j,2*i])      # du
#            o[2*i,2*j+1] = np.conjugate(d[2*j+1,2*i])  # uu
#            o[2*i+1,2*j] = np.conjugate(d[2*j,2*i+1])  # dd
#    return o


def enforce_eh_symmetry_anomalous_sector(d01):
    """Enforce electron-hole symmetry in one of the sectors"""
    out01 = dict()
    for key in d01: # loop over keys
        key2 = tuple(-np.array(key)) # minus one
        o01 = d01[key]*0.0
        o01 = enforce_eh_symmetry_anomalous_jit(d01[key],d01[key2],o01)
        out01[key] = o01
    return out01



@jit(nopython=True)
def enforce_eh_symmetry_anomalous_jit(d01,d10,o01):
    """Enforce electron-hole symmetry"""
    ns = len(d01)//2 # number of spinless sites
    for i in range(ns): # loop
        for j in range(ns): # loop
            # enforce the up|up sector
            o01[2*i,2*j+1] = d01[2*i,2*j+1] - d10[2*j,2*i+1]
            # enforce the down|down sector1
            o01[2*i+1,2*j] = d01[2*i+1,2*j] - d10[2*j+1,2*i]
            # enforce the up|down sector1 (beware of the minus sign)
            o01[2*i,2*j] = d01[2*i,2*j] + d10[2*j+1,2*i+1]
            o01[2*i+1,2*j+1] = d01[2*i+1,2*j+1] + d10[2*j,2*i]
    return o01/2.




import numpy as np


def kchain(h,k):
  """ Return the kchain Hamiltonian """
  if h.dimensionality != 2: raise
  if h.is_multicell: h = h.get_no_multicell() # redefine
  tky = h.ty*np.exp(1j*np.pi*2.*k)
  tkxy = h.txy*np.exp(1j*np.pi*2.*k)
  tkxmy = h.txmy*np.exp(-1j*np.pi*2.*k)  # notice the minus sign !!!!
  # chain in the x direction
  ons = h.intra + tky + np.conjugate(tky).T  # intra of k dependent chain
  hop = h.tx + tkxy + tkxmy  # hopping of k-dependent chain
  return (ons,hop)





import numpy as np
import scipy.linalg as lg
from . import algebra
from .ldos import spatial_dos,write_ldos
import os
from . import filesystem as fs
from . import parallel

def evolve_local_state(h,i=0,ts=np.linspace(0.,20.,300),
        mode="chi"):
    """Evolve a state that is originally localized in a point"""
    if h.dimensionality!=0: raise # only for 0d
    # create the vector
#    if h.has_spin: raise
    # get the function that does time evolution
    if mode=="green": evol = evolve_green(h,i=i)
    elif mode=="chi": evol = evolve_chi(h,i=i,ts=ts)
    g = h.geometry
    fs.rmdir("MULTITIMEEVOLUTION") # remove folder
    fs.mkdir("MULTITIMEEVOLUTION") # create folder
    fo = open("MULTITIMEEVOLUTION/MULTITIMEEVOLUTION.TXT","w")
    for t in ts: # loop over ts
        out = evol(t) # do the evolution
        print(np.sum(out))
        out = spatial_dos(h,out) # resum if necessary
        name = "TIMEEVOLUTION_T_"+str(t)+"_.OUT" # name
        name2 = "MULTITIMEEVOLUTION/"+name # name
        write_ldos(g.x,g.y,out,output_file=name2) # write the LDOS
        fo.write(name+"\n") # write this file
    fo.close()




def evolve_green(h,i=0):
    """Return the Green function evolution"""
    if h.dimensionality!=0: raise # only for 0d
    # create the vector
    v0 = np.zeros(h.intra.shape[0]) # zero dimensional
    v0[i] = 1.0
    # now do the time evolution
    (es,vs) = algebra.eigh(h.intra) # diagonalize
    vs = vs.T # eigenvectors
    vs = vs[es<0.0,:] # get the eigenvectors
    es = es[es<0.0]
    ws = np.array([np.conjugate(iv[i]) for iv in vs]) # weights
    def evol(t):
        phi = np.exp(1j*es*t) # complex phases
        out2 = np.array([iw*iv*iphi for (iw,iv,iphi) in zip(ws,vs,phi)])
        out = np.zeros(v0.shape[0]) 
        out += np.abs(np.sum(out2,axis=0))**2 # sum over eigenvectors
        return out
    return evol





def evolve_chi(h,i=0,ts=[0]):
    """Function that return the time evolution for chi"""
    from .chi import chargechi_row
    emin = np.min(algebra.eigvalsh(h.intra))
    es = np.linspace(emin,0.,int(max(ts)*2))
    cs = chargechi_row(h,es=es,i=i,delta=abs(emin)/len(es))
    def evol(t):
        out = np.array([c@np.exp(1j*es*t) for c in cs])
        return np.abs(out)
    return evol



import numpy as np

# functions to compute the effective mass


def mass_operator(h,**kwargs):
    """Return the operator to compute the effective mass"""
    def mop(v,k=None):
        if k is None: 
            print("Kpoint required")
            raise
        hk0 = h.get_hk_gen()(k) # get the Hamiltonian
        from .algebra import eigh # function to diagonalize
        e0,w0 = eigh(hk0) # central point 
        m = effective_mass_velocity(h,k,w0=w0,**kwargs) # array with masses
        T = np.matrix(w0) # convert
        m0 = np.diag(m) # build new hamiltonian
        #O = np.array(np.conjugate(T)@m0@T.T) # operator
        #O = np.conjugate(np.array(T@m0@T.H)) # operator
        O = np.array(T@m0@T.H) # operator
        out =  O@v # return the result
        return out
    from .operators import Operator
    return Operator(mop)


def effective_mass(h,k,dk=1e-2,**kwargs):
    """Given a Hamiltonian and a k-point, return the effective mass"""
    from .topology import smooth_gauge # perform rotation if needed
    from .algebra import eigvalsh # function to diagonalize
    if h.dimensionality==1: # not implemented
        k = np.array(k)
        hk = h.get_hk_gen() # get generator
        dkv = np.array([dk,0.,0.])
        e0 = np.sort(eigvalsh(hk(k)))
        ep = np.sort(eigvalsh(hk(k+dk)))
        em = np.sort(eigvalsh(hk(k-dk)))
        m = ((e0 - em) - (ep - e0))/(dk*dk) # array with effective mass
        a1 = h.geometry.a1
        v = np.sqrt(a1.dot(a1))
        m = v*m/((2.*np.pi)**2) # renormalize by the pi factors
        return m
    else: raise




def effective_mass_velocity(h,k,dk=1e-3,w0=None):
    """Given a Hamiltonian and a k-point, return the effective mass"""
    from .algebra import eigh,eigvalsh # function to diagonalize
    from .algebra import braket_wAw # function to diagonalize
#    if not h.dimensionality==1: raise # not implemented
    k = np.array(k)
    hk = h.get_hk_gen() # get generator
    dkv = np.array([dk,0.,0.])
    hk0 = hk(k) # Hamiltonian
    if w0 is None: # initial waves not provided
        e0,w0 = eigh(hk0) # central point 
        w0 = w0.T.copy() # transpose
    else:
        w0 = w0.T.copy() # transpose
        e0 = np.array([braket_wAw(w,hk0) for w in w0]) # w0 is given
    from .current import current_operator
#    J = current_operator(h) # current generator
    from .current import derivative
    def d2edk2(i,j): # compute derivative
        if i==0 and j==0: order = [2,0,0]
        if i==0 and j==1: order = [1,1,0]
        if i==1 and j==0: order = [1,1,0]
        if i==1 and j==1: order = [0,2,0]
        mass = lambda k: derivative(h,k,order=order)
#        vm = np.array([braket_wAw(w,J(k-dkv)) for w in w0]) # velocity before
#        vp = np.array([braket_wAw(w,J(k+dkv)) for w in w0]) # velocity before
#        m = (vp-vm)/dk
        m = np.array([braket_wAw(w,mass(k)) for w in w0]) # velocity before
        return m
    d = h.dimensionality
    m = np.zeros((len(w0),d,d),dtype=np.complex_)
    for i in range(d):
      for j in range(d):
        m[:,i,j] = d2edk2(i,j) # get all the masses
    m = np.array([np.trace(m[i,:,:]) for i in range(len(w0))]) # eigenmass
#    m = [eigvalsh(m[i,:,:]) for i in range(len(w0))] # eigenmass
#    m = np.array([np.sum(mi) for mi in m]) # sum of the mass
    a1 = h.geometry.a1
    v = np.sqrt(a1.dot(a1))
    m = v*m/((2.*np.pi)**2) # renormalize by the pi factors
#    return np.sign(m)
    return m # return the array with masses in the diagonal basis





from . import multicell
from . import sculpt
from . import sculpt

def build(h,nz=1):
  """Create Hamiltonian of a film from a 3d geometry"""
#  if not h.dimensionality==3: raise
  ho = multicell.supercell(h,nsuper=[1,1,nz],sparse=False,ncut=3)
  ho.dimensionality = 2 # reduce dimensionality
  ho.geometry.dimensionality = 2 # reduce dimensionality
  ho.geometry = sculpt.set_xy_plane(ho.geometry) # put in the xy plane
  hoppings = [] # empty list
  for t in ho.hopping: # loop over hoppings
    if t.dir[2]== 0: hoppings.append(t) # remove in the z direction
  ho.hopping = hoppings # overwrite hoppings
  return ho



def geometry_film(g,nz=1):
  """Create the geometry of a film"""
  g = sculpt.set_xy_plane(g) # put in the xy plane
  go = g.get_supercell([1,1,nz]) # create the supercell
  go.dimensionality = 2 # reduce dimensionality
  go.get_fractional(center=True)
  go.fractional2real()
  return go




def hamiltonian_film(hin,nz=10):
  """Return the Hamiltonian of a film"""
  h = hin.copy() # copy Hamiltonian
  h = multicell.supercell_hamiltonian(h,nsuper=[1,1,nz])
  hopout = [] # list
  for i in range(len(h.hopping)): # loop over hoppings
    if abs(h.hopping[i].dir[2])<0.1: 
      hopout.append(h.hopping[i])
  h.hopping = hopout
  h.dimensionality = 2
  h.geometry.dimensionality = 2
  h.geometry = sculpt.set_xy_plane(h.geometry)
  return h





# library to create multilayer systems
from copy import deepcopy
import numpy as np
from scipy.sparse import csc_matrix as csc
from scipy.sparse import csc_matrix 
from scipy.sparse import bmat
from . import geometry


def honeycomb_ribbon(n=1,rtype="zigzag",stype="AB",has_spin=True,
                     t = 0.1,efield=0.1):
  """Return the hamiltonian of a ribbon"""
  if rtype=="zigzag":  g = geometry.honeycomb_zigzag_ribbon(n)
  elif rtype=="armchair":  g = geometry.honeycomb_zigzag_ribbon(n)
  else: raise
  if rtype=="amrchair": raise # not implemented yet
  h = g.get_hamiltonian(has_spin=has_spin) # hamiltonian
  if stype=="ABC": # trilayer ABC
    h = build_honeycomb_trilayer(h,t,mvl=[0.,1.])
  elif stype=="AB": # trilayer ABC
    h = build_honeycomb_bilayer(h,t,mvl=[0.,1.])
  add_electric_field(h,e=efield)
  return h




def bilayer_aa(h,t = 0.1):
  """ Creates a bilayer from a honeycomb ribbon"""
  nlayers = 2 # number of layers
  g = h.geometry # get the geometry
  go = deepcopy(g) # copy the geometry
  go.x = [] 
  go.y = []
  go.z = []
  if g.name == "honeycomb_armchair_ribbon": dx ,dy = 1. ,0.
  if g.name == "honeycomb_zigzag_ribbon": dx ,dy = 0. ,-1.
  for (xi,yi) in zip(g.x,g.y):  # modify the geometry
    go.x.append(xi)
    go.x.append(xi+dx)
    go.y.append(yi)
    go.y.append(yi+dy)
    go.z.append(1.)
    go.z.append(-1.)
  go.x,go.y,go.z = np.array(go.x),np.array(go.y),np.array(go.z) # put arrays
  # now modify the hamiltonian
  ho = deepcopy(h)
  n = len(ho.intra) # dimension
  intra = [[0. for i in range(2*n)] for j in range(2*n)]
  inter = [[0. for i in range(2*n)] for j in range(2*n)]
  norb = n # number of orbitals
  # get the atoms which hop according to monolayer type...
  if h.has_spin: 
    norb = norb/2
    tl = [] # interlayer pairs
    x, y, z = go.x, go.y, go.z
    for i in range(len(x)): # loop over atoms
      for j in range(len(x)): # loop over atoms
        if 1.9 < np.abs(z[i]-z[j]) < 2.1: # if in contiguous layers 
          dd = (x[i]-x[j])**2 + (y[i] - y[j])**2 + (z[i] - z[j])**2
          if 3.9<dd<4.1:
            tl.append([i,j])
    for i in range(norb):
      for j in range(norb):  # assign interlayer hopping
        for s in range(2):
         for l in range(nlayers):
          intra[2*nlayers*i+s+2*l][2*nlayers*j+s+2*l] = h.intra[2*i+s,2*j+s]
          inter[2*nlayers*i+s+2*l][2*nlayers*j+s+2*l] = h.inter[2*i+s,2*j+s]
    # now put the interlayer hopping
    for p in tl:
      for s in range(2): # loop over spin
        intra[2*p[0]+s][2*p[1]+s] = t  
        intra[2*p[1]+s][2*p[0]+s] = np.conjugate(t)  
  else: raise # not implemented...
  if h.has_eh: raise # not implemented ....
  ho.intra = np.matrix(intra)
  ho.inter = np.matrix(inter)
  ho.geometry = go
  return ho   



def add_electric_field(h,e = 0.0):
  """Adds electric field to the system"""
  h.add_onsite(lambda r: e*r[2])
  return h

def multilayered_hamiltonian(h,dr=np.array([0.,0.,0.])):
  """ Creates a multilayered hamiltonian by adding several layers """






def add_interlayer(t,ri,rj,has_spin=True,is_sparse=True):
  """Calculate interlayer coupling"""
  m = np.matrix([[0. for i in ri] for j in rj])
  if has_spin: m = bmat([[csc(m),None],[None,csc(m)]]).todense()
  zi = [r[2] for r in ri]
  zj = [r[2] for r in rj]
  for i in range(len(ri)): # add the interlayer
    for j in range(len(ri)): # add the interlayer
      rij = ri[i] - rj[j] # distance between atoms
      dz = zi[i] - zj[j] # vertical distance
      if (3.99<rij.dot(rij)<4.01) and (3.99<(dz*dz)<4.01): # check if connect
        if has_spin: # spin polarized
          m[2*i,2*j] = t
          m[2*i+1,2*j+1] = t
        else:  # spin unpolarized
          m[i,j] = t
  return m







def build_honeycomb_bilayer(h,t=0.1,mvl = None ):
  """Builds a multilayer based on a hamiltonian, it is assumed that
  there are no new intercell hoppings in the multilayer"""
  g = h.geometry  # save geometry
  ho = deepcopy(h) # copy the hamiltonian
  go = deepcopy(g) # copy the geometry
  if mvl is None: # if not provided assume firs neighbors
    mvl = g.r[0] - g.r[1]
  def mono2bi(m):
    """Increase the size of the matrices"""    
    if h.is_sparse:
      return bmat([[csc(m),None],[None,csc(m)]])
    else:
      return bmat([[csc(m),None],[None,csc(m)]]).todense()  
  # modify the geometry
  x,y,z = np.array(g.x), np.array(g.y) ,np.array(g.z) # store positions
  go.x = np.concatenate([x,x+mvl[0]])  # move 
  go.y = np.concatenate([y,y+mvl[1]])  # move
  go.z = np.concatenate([z-1.0,z+1.0])  # separate two units
  go.xyz2r() # update r coordinates
  if g.has_sublattice: # if has sublattice, keep the indexes
    go.sublattice = g.sublattice*2 # two times
  ho.geometry = go # update geometry
  # modify the hamiltonian
  ho.intra = mono2bi(h.intra)  # increase intracell matrix
  ho.intra = ho.intra + add_interlayer(t,go.r,go.r,has_spin=h.has_spin) # add interlayer coupling
  if h.dimensionality==2: # two dimensional system
    ho.tx = mono2bi(h.tx)  # increase intracell matrix
    ho.ty = mono2bi(h.ty)  # increase intracell matrix
    ho.txy = mono2bi(h.txy)  # increase intracell matrix
    ho.txmy = mono2bi(h.txmy)  # increase intracell matrix
  elif h.dimensionality==1: # one dimensional system
    ho.inter = mono2bi(h.inter)  # increase intercell matrix
    dx = g.a1
    ho.inter += add_interlayer(t,go.r,go.r+dx) # add interlayer coupling
  else:
    raise
  return ho



def build_honeycomb_trilayer(h,t=0.1,mvl=None):
  """Builds a multilayer based on a hamiltonian, it is assumed that
  there are no new intercell hoppings in the multilayer"""
  g = h.geometry  # save geometry
  ho = deepcopy(h) # copy the hamiltonian
  go = deepcopy(g) # copy the geometry
  if mvl is None: # if not provided assume firs neighbors
    mvl = g.r[0] - g.r[1]
  def mono2tri(m):
    """Increase the size of the matrices"""    
    mo = [[None for i in range(3)] for j in range(3)]
    for i in range(3): 
      mo[i][i] = csc(m)
    return bmat(mo).todense()
  # modify the geometry
  x,y,z = np.array(g.x), np.array(g.y) ,np.array(g.z) # store positions
  go.x = np.concatenate([x-mvl[0],x,x+mvl[0]])  # move one unit to the right
  go.y = np.concatenate([y-mvl[1],y,y+mvl[1]])  # leave invariant
  go.z = np.concatenate([z-2.0,z,z+2.0])  # separate two units
  go.xyz2r() # update r coordinates
  if g.has_sublattice: # if has sublattice, keep the indexes
    go.sublattice = g.sublattice*3 # three times
  ho.geometry = go # update geometry
  # modify the hamiltonian
  ho.intra = mono2tri(h.intra)  # increase intracell matrix
  ho.intra += add_interlayer(t,go.r,go.r) # add interlayer coupling
  if h.dimensionality==2: # two dimensional system
    ho.tx = mono2tri(h.tx)  # increase intracell matrix
    ho.ty = mono2tri(h.ty)  # increase intracell matrix
    ho.txy = mono2tri(h.txy)  # increase intracell matrix
    ho.txmy = mono2tri(h.txmy)  # increase intracell matrix
  elif h.dimensionality==1: # one dimensional system
    ho.inter = mono2tri(h.inter)  # increase intercell matrix
    dx = g.a1
    ho.inter += add_interlayer(t,go.r,go.r+dx) # add interlayer coupling
  else:
    raise
  ## add sublattice index, might break
  return ho




def get_armchair_bilayer(n=10):
  """Get hamiltonian of an armchair bilayer"""
  g = geometry.honeycomb_armchair_ribbon(30)
  drs = [np.array([0.,0.,-1.]),np.array([np.sqrt(3.)/2.,0.,1.])]
  g = geometry.apilate(g,drs=drs)
  h = g.get_hamiltonian()
  return h


def get_zigzag_bilayer(n=10):
  """Get hamiltonian of a zigzag bilayer"""
  g = geometry.honeycomb_zigzag_ribbon(40)
  drs = [np.array([0.,0.,-1.]),np.array([0.,1.,1.])]
  g = geometry.apilate(g,drs=drs)
  h = g.get_hamiltonian()
  return h



def bilayer_geometry(g,mvl=None,dz=2.0):
  if mvl is None: # if not provided assume firs neighbors
    mvl = g.r[0] - g.r[1]
  go = g.copy() # copy geometry
  x,y,z = np.array(g.x), np.array(g.y) ,np.array(g.z) # store positions
  go.x = np.concatenate([x,x+mvl[0]])  # move 
  go.y = np.concatenate([y,y+mvl[1]])  # move
  go.z = np.concatenate([z-dz/2,z+dz/2])  # separate two units
  go.xyz2r() # update r coordinates
  if g.has_sublattice: # if has sublattice, keep the indexes
    sl = g.sublattice
    go.sublattice = np.concatenate([sl,sl]) # two times
  return go



def get_geometry(name,dz=2.0,armchair=True):
  """Return the geometry for multilayer graphene"""
  g = geometry.honeycomb_lattice() # honeycomb lattice
  if name=="AA":
    g = bilayer_geometry(g,mvl=[0.,0.],dz=1.0)
  elif name=="AB":
    g = bilayer_geometry(g,mvl=None,dz=1.0)
  else: raise
  g.has_sublattice = False # no sublattice
  if armchair: g = g.supercell([[1,-1,0],[0,1,0],[0,0,1]]) # armchair unit cell
  g.z *= dz ; g.xyz2r() # increase distance
  return g



def multilayer_hopping(dz=2.0,ti=0.3):
  """Function to calculate the hopping in multilayer systems"""
  def fun(r1,r2):
    dr = r1-r2
    if 0.9<dr.dot(dr)<1.1: return 1.0 # first neighbor
    if abs(dr.dot(dr)-dz**2)<0.1: 
      if abs(dr[2]**2-dz**2)<0.1: 
        return ti # first neighbor
    return 0.0
  return fun




from __future__ import print_function
import os

# names is a lists with pairs of name of folder, f90 file and .so file

def compile_fortran(compiler=None):
    names = [("first_neigh","first_neighborsf90.f90","first_neighborsf90")] 
    names += [("kpm","kpm.f90","kpmf90")] 
    names += [("dos","dos.f90","dosf90")] 
    names += [("berry","berry_curvature.f90","berry_curvaturef90")] 
    names += [("gauss_inv","gauss_inv.f90","gauss_invf90")] 
    names += [("clean_geometry","clean_geometryf90.f90","clean_geometryf90")] 
    names += [("correlators","correlatorsf90.f90","correlatorsf90")]
    names += [("supercell","supercellf90.f90","supercellf90")] 
    names += [("classicalspin","classicalspinf90.f90","classicalspinf90")] 
    names += [("density_matrix","density_matrixf90.f90","density_matrixf90")] 
    names += [("kanemele","kanemelef90.f90","kanemelef90")] 
    names += [("green","greenf90.f90","greenf90")] 
    names += [("specialhopping","specialhoppingf90.f90","specialhoppingf90")] 
    names += [("chi","chif90.f90","chif90")] 
    names += [("tails","tailsf90.f90","tailsf90")] 
    names += [("algebra","algebraf90.f90","algebraf90")] 
    names += [("dyson","dyson2d.f90","dyson2df90")] 
    
    import sys
    import os
    if compiler is None:
      compiler = os.path.dirname(os.path.realpath(sys.executable))+"/f2py"
    if not os.path.isfile(compiler):
        print(compiler,"does not exist")
        print("You may want to use another Python")
        exit()
    
    print("Using compiler",compiler)
    dirname = os.path.dirname(os.path.realpath(__file__)) # main folder
    pwd = os.getcwd() # current directory
    
    for name in names:
      folder,f90,so = name[0],name[1],name[2] # different names
      os.chdir(dirname+"/fortran/"+folder) # go to the folder
      os.system("rm -f *.so") # remove old libraries
      print("Compiling",name[1])
      os.system(compiler+" -c -m "+so+"  "+f90+"> /dev/null 2>&1") # compile
      os.system("cp *.so ../../"+so+".so") # copy library
    os.chdir(pwd) # return to parent directory

if __name__=="__main__":
    compile_fortran()



import numpy as np


def get_k2K_generator(self,toreal=False):
      """
      Function to turn a reciprocal lattice vector to natural units
      """
      R = get_k2K(self) # get matrix
      if toreal: R = np.conjugate(R).T # transform to real coordinates
      def fun(k0):
        if len(k0)==3: k = k0 # do nothing
        elif len(k0)==2: k = np.array([k0[0],k0[1],0.]) # convert to three comp
        r = np.array(k) # real space vectors
        out = np.array(R)@r # change of basis
        if len(k0)==2: return np.array([out[0],out[1]]) # return two
        return out
      return fun




def get_k2K(g):
  """Return a matrix that converts vectors
  in the reciprocal space into natural units, useful for drawing
  2D/3D quantities"""
  if g.dimensionality == 2:
    (ux,uy,uz) = (g.a1,g.a2,np.array([0.,0.,1]))
  elif g.dimensionality == 3:
    (ux,uy,uz) = (g.a1,g.a2,g.a3)
  else: raise
  ux = ux/np.sqrt(ux.dot(ux))
  uy = uy/np.sqrt(uy.dot(uy))
  uz = uz/np.sqrt(uz.dot(uz))
  a2kn = np.matrix([ux,uy,uz]) # matrix for the change of basis
  r2a = np.matrix([ux,uy,uz]).T.I # from real space to lattice vectors
  R = a2kn@r2a@a2kn.T # rotation matrix
  return R


def unitary(v): return v/np.sqrt(v.dot(v))




import numpy as np
from ..superconductivity import time_reversal

def has_time_reversal_symmetry(h):
    """Check if a Hamiltonian breaks time reversal symmetry"""
    if h.has_eh: 
        print("WARNING, time reversal not implemented for BdG")
        return False
    else:
        if h.has_spin: f = time_reversal
        else: f = lambda x: np.conjugate(x)
        h1 = h.copy()
        h1.modify_hamiltonian_matrices(f) # apply time reversal
        d0 = h.get_multihopping()
        d1 = h1.get_multihopping()
        dd = d0 -d1 # difference
        diff = dd.dot(dd) # scalar product
        if diff.real<1e-7: return True
        else: return False





import numpy as np
from numba import jit

def Vijkl(h,n=4,fv=None):
    """Return the Coulomb matrix elements between different orbitals"""
    numw = n
    if h.dimensionality!=0: raise # only for 0D
    if h.has_eh: raise # not implemented
    if h.has_spin: raise # not implemented
    es,ws = h.get_eigenvectors(numw=numw) # get eigenstates
    g = h.geometry # geometry
    if fv is None:
        def fv(r):
            r2 = r.dot(r)
            if r2<0.1: return 0.0 # zero
            else: return 1./np.sqrt(r2) # coulomb
    vs = [[fv(r1-r2) for r1 in g.r] for r2 in g.r] # interaction
    out = np.zeros((numw,numw,numw,numw),dtype=np.complex) # storage
    return get_vijkl_jit(ws,np.array(vs),out)


@jit(nopython=True)
def get_vijkl_jit(ws,vs,out):
    """Return the Vijkl elements, given a distance interaction vs"""
    n = len(ws) # number of wavefunctions
    nc = len(vs) # number of components
    for i in range(n):
        wi = np.conjugate(ws[i])
        for j in range(n):
            wj = ws[j]
            for k in range(n):
                wk = np.conjugate(ws[k])
                for l in range(n):
                    wl = ws[l]
                    o = 0.0j # initialize
                    for ii in range(nc):
                        for jj in range(nc):
                            o = o + vs[ii,jj]*wi[ii]*wj[ii]*wk[jj]*wl[jj]
                    out[i,j,k,l] = o # store
    return out # return output



import numpy as np
from ..green import gauss_inverse # calculate the desired green functions
from ..algebra import dagger,sqrtm
from .smatrix import enlarge_hlist,effective_tridiagonal_hamiltonian
from .smatrix import build_effective_hlist
from ..checkclass import is_iterable

def get_full_green(ht,energy,mode="right",delta=None,ic=0):
    """Build effective Hamiltonian at a certain energy"""
    if delta is None: delta = ht.delta
    # get the selfenergies, using the same coupling as the lead
    selfl = ht.get_selfenergy(energy,delta=delta,lead=0,pristine=True)
    selfr = ht.get_selfenergy(energy,delta=delta,lead=1,pristine=True)
    if ht.block_diagonal:
        ht2 = enlarge_hlist(ht) # get the enlaged hlist with the leads
    # selfenergy of the leads (coupled to another cell of the lead)
        gmatrix = effective_tridiagonal_hamiltonian(ht2.central_intra,
                                        selfl,selfr,
                                        energy=energy,
                                        delta=delta + ht.extra_delta_central)
        test_gauss = False # do the tridiagonal inversion
    else: # not block diagonal
        gmatrix = build_effective_hlist(ht,energy=energy,
                                          delta=delta,selfl=selfl,
                                        selfr=selfr)
    test_gauss = True # do the tridiagonal inversion
    if mode=="left": g = gauss_inverse(gmatrix,0,0,test=test_gauss)
    elif mode=="right": g = gauss_inverse(gmatrix,-1,-1,test=test_gauss)
    elif mode=="central": 
        if ic is None: # provide a list of all the Green's functions
            out = [] # empty list
            for ii in range(len(gmatrix)):
              g = gauss_inverse(gmatrix,i=ii,j=ii,test=test_gauss)
              out.append(g)
            return out
        elif is_iterable(ic): # an iterable  
            out = [] # empty list
            for ii in ic:
              g = gauss_inverse(gmatrix,i=ii+1,j=ii+1,test=test_gauss)
              out.append(g)
            return out
        else: # well implemented way, assume it is a number
          ii = 1+ic
          g = gauss_inverse(gmatrix,i=ii,j=ii,test=test_gauss)
    else: raise
    return g



import numpy as np
from ..green import gauss_inverse # calculate the desired green functions
from ..algebra import dagger,sqrtm

delta_smatrix = 1e-12 # delta for the smatrix



def get_smatrix(ht,energy=0.0,as_matrix=False,check=True):
    """Calculate the S-matrix of an heterostructure"""
    # now do the Fisher Lee trick
    delta = ht.delta
    smatrix = [[None,None],[None,None]] # smatrix in list form
    # get the selfenergies, using the same coupling as the lead
    selfl = ht.get_selfenergy(energy,delta=delta,lead=0,pristine=True)
    selfr = ht.get_selfenergy(energy,delta=delta,lead=1,pristine=True)
    # get the central Green's function
    gmatrix = ht.get_central_gmatrix(selfl=selfl,selfr=selfr,
                                   energy=energy)
    # gamma functions
    test_gauss = True # gauss only works with square matrices
    gammar = 1j*(selfr-dagger(selfr))
    gammal = 1j*(selfl-dagger(selfl))
    # calculate the relevant terms of the Green function
    g11 = gauss_inverse(gmatrix,0,0,test=test_gauss)
    g12 = gauss_inverse(gmatrix,0,-1,test=test_gauss)
    g21 = gauss_inverse(gmatrix,-1,0,test=test_gauss)
    g22 = gauss_inverse(gmatrix,-1,-1,test=test_gauss)
    ######## now build up the s matrix with the fisher trick
    # the identity can have different dimension ignore for now....
    iden = np.matrix(np.identity(g11.shape[0],dtype=complex)) # create identity
    iden11 = np.matrix(np.identity(g11.shape[0],dtype=complex)) # create identity
    iden22 = np.matrix(np.identity(g22.shape[0],dtype=complex)) # create identity
    smatrix[0][0] = -iden + 1j*sqrtm(gammal)@g11@sqrtm(gammal) # matrix
    smatrix[0][1] = 1j*sqrtm(gammal)@g12@sqrtm(gammar) # transmission matrix
    smatrix[1][0] = 1j*sqrtm(gammar)@g21@sqrtm(gammal) # transmission matrix
    smatrix[1][1] = -iden + 1j*sqrtm(gammar)@g22@sqrtm(gammar) # matrix
    if check: # check whether the matrix is unitary
        from .unitarize import check_and_fix
        smatrix = check_and_fix(smatrix,error=100*delta)
    if as_matrix:
      from scipy.sparse import bmat,csc_matrix
      smatrix2 = [[csc_matrix(smatrix[i][j]) for j in range(2)] for i in range(2)]
      smatrix = bmat(smatrix2).todense()
    return smatrix


def get_central_gmatrix(ht,selfl=None,selfr=None,energy=0.0):
    """Return the inverse of the central Green's function"""
    delta = ht.delta
    if selfl is None: selfl = ht.get_selfenergy(energy,
        delta=delta,lead=0,pristine=True)
    if selfr is None: selfr = ht.get_selfenergy(energy,
        delta=delta,lead=1,pristine=True)
    if ht.block_diagonal:
        ht2 = enlarge_hlist(ht) # get the enlaged hlist with the leads
        gmatrix = effective_tridiagonal_hamiltonian(ht2.central_intra,
                                      selfl,selfr,
                                      energy=energy,
                                      delta=delta + ht.extra_delta_central)
    else: # not block diagonal
        gmatrix = build_effective_hlist(ht,energy=energy,delta=delta,
                                       selfl=selfl,
                                      selfr=selfr)
    return gmatrix




def effective_tridiagonal_hamiltonian(intra,selfl,selfr,
                                        energy = 0.0, delta=1e-5):
    """ Calculate effective Hamiltonian"""
    if not type(intra) is list: raise # assume is list
    n = len(intra) # number of blocks
    iout = [[None for i in range(n)] for j in range(n)] # empty list
    iden = np.matrix(np.identity(intra[0][0].shape[0],dtype=np.complex_))
    if delta>delta_smatrix: delta = delta_smatrix # small delta is critical!
    ez = iden*(energy +1j*delta) # complex energy
    for i in range(n):
      iout[i][i] = ez - intra[i][i] # simply E -H
    for i in range(n-1):
      iout[i][i+1] = -intra[i][i+1] # simply E -H
      iout[i+1][i] = -intra[i+1][i] # simply E -H
    # and now the selfenergies
    iout[0][0] = iout[0][0] -selfl
    iout[-1][-1] = iout[-1][-1] -selfr
    return iout




def enlarge_hlist(ht):
    """Add a single cell of the leads to the central part"""
    ho = ht.copy() # copy heterostructure
    if not ht.block_diagonal: raise # check that is in block diagonal form
    nc = len(ht.central_intra) # number of cells in the central
    hcentral = [[None for i in range(nc+2)] for j in range(nc+2)]
    for i in range(nc): # intraterm
      hcentral[i+1][i+1] = ht.central_intra[i][i].copy() # common
    for i in range(nc-1): # interterm
      hcentral[i+1][i+2] = ht.central_intra[i][i+1].copy() # common
      hcentral[i+2][i+1] = ht.central_intra[i+1][i].copy() # common
    # now the new terms
    hcentral[0][0] = ht.left_intra.copy() # left
    hcentral[-1][-1] = ht.right_intra.copy() # right
    if nc>0: # more than two cells in the center
        hcentral[0][1] = dagger(ht.left_coupling)*ht.scale_lc # left
        hcentral[1][0] = ht.left_coupling*ht.scale_lc # left
        hcentral[-2][-1] = ht.right_coupling*ht.scale_rc # right
        hcentral[-1][-2] = dagger(ht.right_coupling)*ht.scale_rc # right
    else: # no original central part
        # here the average of the two hoppings will be performed
        # if the Hamiltonian of the scattering region is not provided
        # this may be not the optimal intuitive choice
        # perhaps square root of the product is more natural
        # gamma = sqrtm(dagger(ht.left_coupling)@ht.right_coupling)?
        hcentral[0][1] = dagger(ht.left_coupling) # left
        hcentral[1][0] = ht.left_coupling.copy() # left
        hcentral[-2][-1] += ht.right_coupling # right
        hcentral[-1][-2] += dagger(ht.right_coupling) # right
        hcentral[0][1] *= ht.scale_rc*ht.scale_lc/2. # factor 1/2 for DC
        hcentral[1][0] *= ht.scale_rc*ht.scale_lc/2. # factor 1/2 for DC
    # store in the object
    ho.central_intra = hcentral
    # and redefine the new lead couplings
    ho.right_coupling = ht.right_inter
    ho.left_coupling = ht.left_inter
    return ho # return the new heterostructure











def build_effective_hlist(ht,energy=0.0,delta=0.0001,selfl=None,selfr=None):
    """ Calculate list of effective Hamiltonian which will be inverted"""
    if (selfl is None) or (selfr is None):
        (selfl,selfr) = get_surface_selfenergies(ht,energy=energy,delta=delta,
                        pristine=True)
    intra = ht.central_intra # central intracell hamiltonian
    if len(ht.central_intra)==0: # no central part provided
        ht.central_intra = (ht.left_intra + ht.right_intra)/2.
        print("Generating a dummy central cell, you may want to use a different geometry")
    ce = energy +1j*delta
    idenc = np.matrix(np.identity(len(ht.central_intra),dtype=complex))*ce
    idenl = np.matrix(np.identity(len(ht.left_intra),dtype=complex))*ce
    idenr = np.matrix(np.identity(len(ht.right_intra),dtype=complex))*ce
    hlist = [[None for i in range(3)] for j in range(3)] # list of matrices
    # set up the different elements
    # first the intra terms
    hlist[0][0] = idenl - ht.left_intra - selfl
    hlist[1][1] = idenc - ht.central_intra
    hlist[2][2] = idenr - ht.right_intra - selfr
    # now the inter cell
    hlist[0][1] = -dagger(ht.left_coupling)*ht.scale_lc
    hlist[1][0] = -ht.left_coupling*ht.scale_lc
    hlist[2][1] = -dagger(ht.right_coupling)*ht.scale_rc
    hlist[1][2] = -ht.right_coupling*ht.scale_rc
#    for h in hlist: print(h)
    return hlist




import collections
import numpy as np
from collections.abc import Iterable


# function to check if inputs are of certain type

def is_iterable(e): return isinstance(e,Iterable)

def number2array(n,d=3):
    """Given a certain object, return an array"""
    if is_iterable(n): return [ni for ni in n]
    else: return [n for i in range(d)]



from __future__ import print_function
import numpy as np
import scipy.linalg as lg
from .algebra import smalleig # arpack diagonalization
import time
from . import timing
from . import kpm
from . import checkclass
from . import green
from . import algebra
from . import parallel
from numba import jit
from .klist import kmesh

def calculate_dos(es,xs,d,w=None):
    if w is None: w = np.zeros(len(es)) + 1.0 # initialize
    else: w = w.real # make it real just in case
    es = np.array(es)
    xs = np.array(xs)
    d = np.array(d)
    ys = np.zeros(np.array(xs.shape[0])) # initialize
    ys = calculate_dos_jit(es,xs,d,w,ys) # compute
    return ys

@jit(nopython=True)
def calculate_dos_jit(es,xs,d,w,ys):
      for i in range(len(es)): # loop over energies
          e = es[i]
          iw = w[i]
          de = np.abs(xs - e) # E - Ei
          de = d/(d*d + de*de) # 1/(delta^2 + (E-Ei)^2)
          ys += de*iw # times the weight
      return ys




def dos_surface(h,output_file="DOS.OUT",
                 energies=np.linspace(-1.,1.,20),delta=0.001):
  """Calculates the DOS of a surface, and writes in file"""
  if h.dimensionality!=1: raise # only for 1d
  fo = open(output_file,"w")
  fo.write("# energy, DOS surface, DOS bulk\n")
  for e in energies: # loop over energies
    print("Done",e)
    gb,gs = green.green_renormalization(h.intra,h.inter,energy=e,delta=delta)
    gb = -np.trace(gb).imag
    gs = -np.trace(gs).imag
    fo.write(str(e)+"     "+str(gs)+"    "+str(gb)+"\n")
  fo.close()




def dos0d(h,energies=np.linspace(-4,4,500),delta=0.01):
  """Calculate density of states of a 0d system"""
  hkgen = h.get_hk_gen() # get generator
  calculate_dos_hkgen(hkgen,[0],
            delta=delta,energies=energies) # conventiona algorithm



def dos0d_kpm(h,use_kpm=True,scale=10,npol=100,ntries=100,fun=None):
  """ Calculate density of states of a 1d system"""
  if h.dimensionality!=0: raise # only for 0d
  if not use_kpm: raise # only using KPM
  h.turn_sparse() # turn the hamiltonian sparse
  mus = np.array([0.0j for i in range(2*npol)]) # initialize polynomials
  mus = kpm.random_trace(h.intra/scale,ntries=ntries,n=npol,fun=fun)
  xs = np.linspace(-0.9,0.9,4*npol) # x points
  ys = kpm.generate_profile(mus,xs) # generate the profile
  write_dos(xs*scale,ys) # write in file


def dos0d_sites(h,sites=[0],scale=10.,npol=500,ewindow=None,refine_e=1.0):
  """ Calculate density of states of a 1d system for a certain orbitals"""
  if h.dimensionality!=0: raise # only for 1d
  h.turn_sparse() # turn the hamiltonian sparse
  mus = np.array([0.0j for i in range(2*npol)]) # initialize polynomials
  hk = h.intra # hamiltonian
  for isite in sites:
    mus += kpm.local_dos(hk/scale,i=isite,n=npol)
  if ewindow is None:  xs = np.linspace(-0.9,0.9,int(npol*refine_e)) # x points
  else:  xs = np.linspace(-ewindow/scale,ewindow/scale,npol) # x points
  ys = kpm.generate_profile(mus,xs) # generate the profile
  write_dos(xs*scale,ys) # write in file











def write_dos(es,ds,output_file="DOS.OUT"):
  """ Write DOS in a file"""
  f = open(output_file,"w")
  for (e,d) in zip(es,ds):
    f.write(str(e)+"     ")
    f.write(str(d.real)+"\n")
  f.close()




def dos1d(h,use_kpm=False,scale=10.,nk=100,npol=100,ntries=2,
          ndos=1000,delta=0.01,ewindow=None,frand=None,
          energies=None):
  """ Calculate density of states of a 1d system"""
  if h.dimensionality!=1: raise # only for 1d
  ks = np.linspace(0.,1.,nk,endpoint=False) # number of kpoints
  if not use_kpm: # conventional method
    hkgen = h.get_hk_gen() # get generator
#    delta = 16./(nk*h.intra.shape[0]) # smoothing
    calculate_dos_hkgen(hkgen,ks,
            delta=delta,energies=energies) # conventiona algorithm
    return np.genfromtxt("DOS.OUT").transpose()
  else:
    h.turn_sparse() # turn the hamiltonian sparse
    hkgen = h.get_hk_gen() # get generator
    yt = np.zeros(ndos) # number of dos
    ts = timing.Testimator("DOS") 
    for i in range(nk): # loop over kpoints
      k = np.random.random(3) # random k-point
      hk = hkgen(k) # hamiltonian
      (xs,yi) = kpm.tdos(hk,scale=scale,npol=npol,frand=frand,ewindow=ewindow,
                  ntries=ntries,ne=ndos)
      yt += yi # Add contribution
      ts.remaining(i,nk)
    yt /= nk # normalize
    write_dos(xs,yt) # write in file
    return xs,yt




def dos1d_sites(h,sites=[0],scale=10.,nk=100,npol=100,info=False,ewindow=None):
  """ Calculate density of states of a 1d system for a certain orbitals"""
  if h.dimensionality!=1: raise # only for 1d
  ks = np.linspace(0.,1.,nk,endpoint=False) # number of kpoints
  h.turn_sparse() # turn the hamiltonian sparse
  hkgen = h.get_hk_gen() # get generator
  mus = np.array([0.0j for i in range(2*npol)]) # initialize polynomials
  for k in ks: # loop over kpoints
    hk = hkgen(k) # hamiltonian
    for isite in sites:
      mus += kpm.local_dos(hk/scale,i=isite,n=npol)
    if info: print("Done",k)
  mus /= nk # normalize by the number of kpoints
  if ewindow is None:  xs = np.linspace(-0.9,0.9,npol) # x points
  else:  xs = np.linspace(-ewindow/scale,ewindow/scale,npol) # x points
  ys = kpm.generate_profile(mus,xs) # generate the profile
  write_dos(xs*scale,ys) # write in file


def calculate_dos_hkgen(hkgen,ks,ndos=100,delta=None,
         is_sparse=False,numw=10,window=None,energies=None):
  """Calculate density of states using the ks given on input"""
  if not is_sparse: # if not is sparse
      m = hkgen([0,0,0]) # get the matrix
      if algebra.issparse(m): 
          print("Hamiltonian is sparse, selecting sparse mode in DOS")
          is_sparse = True
  es = np.zeros((len(ks),hkgen(ks[0]).shape[0])) # empty list
  tr = timing.Testimator("DOS",maxite=len(ks))
  if delta is None: delta = 5./len(ks) # automatic delta
  from . import parallel
  def fun(k): # function to execute
    if parallel.cores==1: tr.iterate() # print the info
    hk = hkgen(k) # Hamiltonian
    t0 = time.perf_counter() # time
    if is_sparse: # sparse Hamiltonian 
      es = algebra.smalleig(hk,numw=numw,tol=delta/1e3) # eigenvalues
      ws = np.zeros(es.shape[0])+1.0 # weight
    else: # dense Hamiltonian
      es = algebra.eigvalsh(hk) # get eigenvalues
      ws = np.zeros(es.shape[0])+1.0 # weight
    return es # return energies
#  for ik in range(len(ks)):  
  out = parallel.pcall(fun,ks) # launch all the processes
  es = [] # empty list
  for o in out: 
      es = np.concatenate([es,o]) # concatenate
#    tr.remaining(ik,len(ks))
#  es = es.reshape(len(es)*len(es[0])) # 1d array
  es = np.array(es) # convert to array
  nk = len(ks) # number of kpoints
  if energies is not None: # energies given on input
    xs = energies
  else:
    if window is None:
      xs = np.linspace(np.min(es)-.5,np.max(es)+.5,ndos) # create x
    else:
      xs = np.linspace(-window,window,ndos) # create x
  ys = calculate_dos(es,xs,delta) # use the Fortran routine
  ys /= nk # normalize by the number of k-points
  ys *= 1./np.pi # normalization of the Lorentzian
  write_dos(xs,ys) # write in file
  print("\nDOS finished")
  return (xs,ys) # return result



def dos_kmesh(h,nk=100,delta=None,random=False,ks=None,
        write=True,
        energies=np.linspace(-1,1,200),**kwargs):
    """Compute the DOS in a k-mesh by using the bandstructure function"""
    if ks is None:  ks = kmesh(h.dimensionality,nk=nk)
    if delta is None: delta = 5./nk # not provided
    if random: ks = [np.random.random(3) for k in ks]
    # compute band structure
    out = h.get_bands(kpath=ks,write=False,**kwargs) 
    if len(out)==2: w = None # use no weight
    else: w = out[2] # use weight of the bands
    ys = calculate_dos(out[1],energies,delta,w=w)/len(ks)
    ys *= 1./np.pi # normalization of the Lorentzian
    if write: write_dos(energies,ys) # write in file
    return (energies,ys) # return result









def dos2d(h,use_kpm=False,scale=10.,nk=100,ntries=1,delta=None,
          ndos=2000,random=True,kpm_window=1.0,
          window=None,energies=None,**kwargs):
  """ Calculate density of states of a 2d system"""
  if h.dimensionality!=2: raise # only for 2d
  ks = []
  from .klist import kmesh
  ks = kmesh(h.dimensionality,nk=nk)
  if random:
    ks = [np.random.random(2) for ik in ks]
    print("Random k-mesh")
  if not use_kpm: # conventional method
    hkgen = h.get_hk_gen() # get generator
    if delta is None: delta = 6./nk
# conventiona algorithm
    (xs,ys) = calculate_dos_hkgen(hkgen,ks,ndos=ndos,delta=delta,
                          is_sparse=h.is_sparse,window=window,
                          energies=energies,**kwargs) 
    write_dos(xs,ys) # write in file
    return (xs,ys)
  else: # use the kpm
    if delta is not None: npol = int(20*scale/delta)
    else: npol = ndos//10
    h.turn_sparse() # turn the hamiltonian sparse
    hkgen = h.get_hk_gen() # get generator
    mus = np.array([0.0j for i in range(2*npol)]) # initialize polynomials
    tr = timing.Testimator("DOS")
    ik = 0
    from . import parallel
    if parallel.cores==1: # serial run
      for k in ks: # loop over kpoints
        ik += 1
        tr.remaining(ik,len(ks))      
        if random: 
          kr = np.random.random(2)
          print("Random sampling in DOS")
          hk = hkgen(kr) # hamiltonian
        else: hk = hkgen(k) # hamiltonian
        mus += kpm.random_trace(hk/scale,ntries=ntries,n=npol)
    else:
        ff = lambda k: kpm.random_trace(hkgen(k)/scale,ntries=ntries,n=npol)
        mus = parallel.pcall(ff,ks) # parallel call
        mus = np.array(mus).sum(axis=0) # sum all the contributions
    mus /= len(ks) # normalize by the number of kpoints
    if energies is None:
      xs = np.linspace(-0.9,0.9,ndos)*kpm_window # x points
    else:  xs = energies/scale
    ys = kpm.generate_profile(mus,xs) # generate the profile
    ys /= scale # rescale
    write_dos(xs*scale,ys) # write in file
    return (xs,ys)




def dos3d(h,scale=10.,nk=20,delta=None,ndos=100,
        random=False,energies=None):
  """ Calculate density of states of a 2d system"""
  if h.dimensionality!=3: raise # only for 2d
  ks = [np.random.random(3) for i in range(nk)] # number of kpoints
  hkgen = h.get_hk_gen() # get generator
  if delta is None: delta = 10./ndos # smoothing
  return calculate_dos_hkgen(hkgen,ks,ndos=ndos,delta=delta,energies=energies) 











def dos2d_ewindow(h,energies=np.linspace(-1.,1.,30),delta=None,info=False,
                    use_green=True,nk=300,mode="adaptive"):
  """Calculate the density of states in certain eenrgy window"""
  ys = [] # density of states
  if delta is None: # pick a good delta value
    delta = 0.1*(max(energies) - min(energies))/len(energies)
  if use_green:
    from .green import bloch_selfenergy
    for energy in energies:
      (g,selfe) = bloch_selfenergy(h,nk=nk,energy=energy, delta=delta,
                   mode=mode)
      ys.append(-g.trace()[0,0].imag)
      if info: print("Done",energy)
    write_dos(energies,ys) # write in file
    return
  else: # do not use green function    
    import scipy.linalg as lg
    kxs = np.linspace(0.,1.,nk)
    kys = np.linspace(0.,1.,nk)
    hkgen= h.get_hk_gen() # get hamiltonian generator
    ys = energies*0.
    weight = 1./(nk*nk)
    for ix in kxs:
      for iy in kys:
        k = np.array([ix,iy,0.]) # create kpoint
        hk = hkgen(k) # get hk hamiltonian
        evals = lg.eigvalsh(hk) # get eigenvalues
        ys += weight*calculate_dos(evals,energies,delta) # add this contribution
      if info: print("Done",ix)
    write_dos(energies,ys) # write in file
    return






def dos1d_ewindow(h,energies=np.linspace(-1.,1.,30),delta=None,info=False,
                    use_green=True,nk=300):
  """Calculate the density of states in certain energy window"""
  ys = [] # density of states
  if delta is None: # pick a good delta value
    delta = 0.1*(max(energies) - min(energies))/len(energies)
  if True: # do not use green function    
    import scipy.linalg as lg
    kxs = np.linspace(0.,1.,nk)
    hkgen= h.get_hk_gen() # get hamiltonian generator
    ys = energies*0.
    weight = 1./(nk)
    for ix in kxs:
      hk = hkgen(ix) # get hk hamiltonian
      evals = lg.eigvalsh(hk) # get eigenvalues
      ys += weight*calculate_dos(evals,energies,delta) # add this contribution
    if info: print("Done",ix)
    write_dos(energies,ys) # write in file
    return











def dos_ewindow(h,energies=np.linspace(-1.,1.,30),delta=None,info=False,
                    use_green=True,nk=300):
  """ Calculate density of states in an energy window"""
  if h.dimensionality==2: # two dimensional
    dos2d_ewindow(h,energies=energies,delta=delta,info=info,
                    use_green=use_green,nk=nk)
  elif h.dimensionality==1: # one dimensional
    dos1d_ewindow(h,energies=energies,delta=delta,info=info,
                    use_green=use_green,nk=nk)
  else: raise # not implemented






def convolve(x,y,delta=None):
  """Add a broadening to a DOS"""
  if delta is None: return y # do nothing
  delta = np.abs(delta) # absolute value
  xnew = np.linspace(-1.,1.,len(y)) # array
  d2 = delta/(np.max(x) - np.min(x)) # effective broadening
  fconv = d2/(xnew**2 + d2**2) # convolving function
  yout = np.convolve(y,fconv,mode="same") # same size
  # ensure the normaliation is the same
  ratio = np.sum(np.abs(y))/np.sum(np.abs(yout))
  yout *= ratio
  return yout # return new array





def dos_kpm(h,scale=10.0,ewindow=4.0,ne=10000,
        delta=0.01,nk=100,operator=None,
        random=True,energies=None,info=False,
        **kwargs):
  """Calculate the KDOS bands using the KPM"""
  operator = h.get_operator(operator)
  if energies is not None: # energies provided
      ewindow = np.max(np.abs(energies)) # true window
      ne = len(energies) # number of energies
      if scale<ewindow: scale = ewindow*2 # redefine
  hkgen = h.get_hk_gen() # get generator
  ks = kmesh(h.dimensionality,nk=nk) # klist
  if random: ks = [np.random.random(3) for k in ks]
  ytot = np.zeros(ne) # initialize
  npol = int(scale/delta) # number of polynomials
  def f(k):
    if info: print("Doing",k)
    hk = hkgen(k) # get Hamiltonian
    if operator is None: op = None # no operator
    else:
        op = operator.get_matrix() # get the matrix of the operator
        if op is None: raise # not implemented
    (x,y) = kpm.pdos(hk,scale=scale,npol=npol,ne=ne,operator=op,
                   ewindow=ewindow,**kwargs) # compute
    return (x,y)
  from . import parallel
  numk = len(ks)
  if parallel.cores==1:
    if info: tr = timing.Testimator("DOS",maxite=numk) # generate object
    for ik in range(len(ks)): # loop over kpoints
      if info: tr.iterate()
      k = ks[ik]
      (x,y) = f(k) # compute
      ytot += y # add contribution
    ytot /= nk # normalize
  else: # parallel calculation
    out = parallel.pcall(f,ks) # compute all
    ytot = np.mean([out[i][1] for i in range(numk)],axis=0) # average DOS
    x = out[0][0] # energies
  if energies is not None: # energies provided
      from scipy.interpolate import interp1d
      finter = interp1d(x,ytot,kind="linear",bounds_error=False,
                    fill_value=0.) # interpolation
      x = energies # redefine x
      ytot = finter(energies) # redefine y
  ytot = ytot*h.intra.shape[0] # by the dimension
  np.savetxt("DOS.OUT",np.array([x,ytot]).T) # save in file
  return (x,ytot)



def get_dos(self,**kwargs):
    """General method to compute the density of states"""
    if self.non_hermitian: # non-Hermitian hamiltonian
        from .nonhermitiantk.nhmethods import get_dos as get_dos_NH
        return get_dos_NH(self,**kwargs)
    else: # Hermitian Hamiltonian
        return get_dos_general(self,**kwargs)



def get_dos_general(h,energies=np.linspace(-4.0,4.0,400),
            use_kpm=False,mode="ED",**kwargs):
  """Calculate the density of states"""
  if use_kpm: # KPM
      ewindow = max([abs(min(energies)),abs(min(energies))]) # window
      return dos_kpm(h,ewindow=ewindow,ne=len(energies),**kwargs)
  else: # conventional methods
      if mode=="ED": # exact diagonalization
          return dos_kmesh(h,energies=energies,**kwargs)
      elif mode in ["Green","RG"]: # Green function formalism
          def fun(e):
              return green.green_operator(h,e=e,**kwargs) 
          ds = parallel.pcall(fun,energies) # compute DOS with an operator
          np.savetxt("DOS.OUT",np.array([energies,ds]).T) # write in a file
          return (energies,ds)
      elif mode=="KPM": 
          return dos_kpm(h,energies=energies,**kwargs)
      else: 
          print("Unrecognized option in DOS")
          raise


dos = get_dos # redefine

def autodos(h,auto=True,**kwargs):
    """Automatic computation of DOS"""
    if auto: # automatic integration
        def f(): return dos(h,**kwargs)[1] # return dos
        x = dos(h,**kwargs)[0] # energies
        from .integration import random_integrate
        y = random_integrate(f)
        np.savetxt("DOS.OUT",np.array([x,y]).T)
        return (x,y)
    return dos(h,**kwargs)




def bulkandsurface(h1,energies=np.linspace(-1.,1.,100),operator=None,
                    delta=0.01,hs=None,nk=30,write=True):
  """Compute the DOS of the bulk and the surface"""
  tr = timing.Testimator("KDOS") # generate object
  ik = 0
  h1 = h1.get_multicell() # multicell Hamiltonian
  kpath = [np.random.random(3) for i in range(nk)] # loop
  dosout = np.zeros((2,len(energies))) # output DOS
  for k in kpath:
    tr.remaining(ik,len(kpath)) # generate object
    ik += 1
    outs = green.surface_multienergy(h1,k=k,energies=energies,delta=delta,hs=hs)
    ie = 0
    for (energy,out) in zip(energies,outs): # loop over energies
      # compute dos
      ig = 0
      for g in out: # loop
        if operator is None: d = -np.trace(g).imag # only the trace 
        elif callable(operator): d = operator(g,k=k) # call the operator
        else:  d = -np.trace(g*operator).imag # assume it is a matrix
        dosout[ig,ie] += d # store
        ig += 1 # increase site
      ie += 1 # increase energy
  # write in file
  dosout/=nk # normalize
  if write:
    np.savetxt("DOS_BULK_SURFACE.OUT",np.array([energies,dosout[0],dosout[1]]).T)
  return energies,dosout[0],dosout[1]




def surface2bulk(h,n=50,nk=3000,delta=1e-3,e=0.0,**kwargs):
    """Compute the DOS from the surface to the bulk"""
    if h.dimensionality==2:
      def f(k):
        (ons,hop) = green.get1dhamiltonian(h,[k,0.,0.])
        gf,sf = green.green_renormalization(ons,hop,energy=e,
                delta=delta,**kwargs)
        out = green.green_surface_cells(sf,hop,ons,delta=delta,e=e,n=n)
        return np.array([-algebra.trace(o).imag for o in out]) # DOS
      ks = np.linspace(0.,1.,nk) # loop
      out = np.mean([f(k) for k in ks],axis=0)
    else: raise
    return np.array([range(n),out]) # return array





def surface_dos(h,energies=None,klist=None,delta=0.01,
                         operator=None):
    bout = [] # empty list, bulk
    sout = [] # empty list, surface
    if klist is None:
        klist = [[i,0.,0.] for i in np.linspace(-.5,.5,50)]
    if energies is None: energies = np.linspace(-.5,.5,50)
    h = h.get_no_multicell()
    def sdos(energy):
        if h.dimensionality==1:
            gs,sf = green.green_renormalization(h.intra,h.inter,
                energy=energy,delta=delta) # surface green function
            return -np.trace(sf).imag # return result
        elif h.dimensionality==2:
            out = 0.0
            for k in klist: # loop over kpoints
                gs,sf = green.green_kchain(h,k=k,energy=energy,delta=delta,
                         only_bulk=False) # surface green function
                out += -np.trace(sf).imag
            return out/len(klist)
        else: raise # not implemented
    return energies,np.array([sdos(e) for e in energies])          




from ..parallel import pcall
import numpy as np
from ..integration import simpson
from scipy.integrate import quad
from .. import parallel



def generic_didv(self,temp=0.,**kwargs):
    """Wrapper to compute the dIdV at finite temperature"""
    if temp==0.: return zero_T_didv(self,**kwargs) # zero temperature
    else: # finite temperature
        return finite_T_didv(self,temp=temp,**kwargs)

from .thermaldidv import finite_T_didv

def zero_T_didv(self,delta=None,**kwargs):
    """Zero temperature dIdV"""
    if delta is None: delta = self.delta # set the own delta
    if self.dimensionality==1: # one dimensional
        return zero_T_didv_1D(self,**kwargs)
    elif self.dimensionality==2: # two dimensional
        return zero_T_didv_2D(self,**kwargs)
    else: raise





def zero_T_didv_1D(self,energy=0.0,delta=None,**kwargs):
    """Wrapper for the dIdV in one dimension"""
    if delta is None: delta = self.delta # set the own delta
    if not self.dimensionality==1: raise # only for one dimensional
    return didv(self,energy=energy,delta=delta,**kwargs)
    try:
        return didv(self,energy=energy,delta=delta,**kwargs)
    except:
        print("Something wrong in didv, returning 0")
        return 1e-10

quadepsrel = 1e-2
quadlimit = 30

def zero_T_didv_2D(self,energy=0.0,delta=None,nk=10,
                   imode="quad",**kwargs):
    """Wrapper for the dIdV in two-dimensions"""
    if delta is None: delta = self.delta # set the own delta
    if not self.dimensionality==2: raise # only for two dimensional
    # function to integrate
    print("Computing",energy)
    f = lambda k: self.generate(k,self.scale_lc,self.scale_rc).didv(energy=energy,delta=delta,**kwargs)
    if imode=="grid":
        out = pcall(f,np.linspace(0.,1.,nk,endpoint=False))
        return np.trapz(out,dx=1./nk)
    elif imode=="simpson":
        return simpson(f,eps=1e-4,xlim=[0.,1.])
    elif imode=="quad":
        return quad(f,0.,1.,epsrel=quadepsrel,limit=quadlimit)[0]
    else: 
        print("Unrecognized imode")
        raise



from .smatrix import get_smatrix
from .. import algebra
dagger = algebra.dagger



def didv(ht,energy=0.0,delta=1e-6,kwant=False,opl=None,opr=None,**kwargs):
    """Calculate differential conductance"""
    if ht.has_eh: # for systems with electons and holes
        s = get_smatrix(ht,energy=energy,check=True) # get the smatrix
        r1,r2 = s[0][0],s[1][1] # get the reflection matrices
        get_eh = ht.get_eh_sector # function to read either electron or hole
        # select the normal lead
        # r1 is normal
        r = ht.get_reflection_normal_lead(s) # return the reflection
        ree = get_eh(r,i=0,j=0) # reflection e-e
        reh = get_eh(r,i=0,j=1) # reflection e-h
        Ree = np.trace(dagger(ree)@ree) # total e-e reflection 
        Reh = np.trace(dagger(reh)@reh) # total e-h reflection 
        G = (ree.shape[0] - Ree + Reh).real # conductance
        return G
    else:
        if kwant:
          if opl is not None or opr is not None: raise # not implemented
          from . import kwantlink
          return kwantlink.transport(ht,energy)
        s = get_smatrix(ht,energy=energy) # get the smatrix
        if opl is not None or opr is not None: # some projector given
          raise # this does not make sense
          U = [[np.identity(s[0][0].shape[0]),None],
                   [None,np.identity(s[1][1].shape[0])]] # projector
          if opl is not None: U[0][0] = opl # assign this matrix
          if opr is not None: U[1][1] = opr # assign this matrix
          #### those formulas are not ok
          s[0][0] = U[0][0]@s[0][0]@U[0][0] # new smatrix
          s[0][1] = U[0][0]@s[0][1]@U[1][1] # new smatrix
          s[1][0] = U[1][1]@s[1][0]@U[0][0] # new smatrix
          s[1][1] = U[1][1]@s[1][1]@U[1][1] # new smatrix
        r1,r2,t = s[0][0],s[1][1],s[0][1] # get the reflection matrices
        # select a normal lead (both of them are)
        # r1 is normal
        ree = r1
        Ree = np.trace(dagger(ree)@ree) # total e-e reflection 
        G1 = (ree.shape[0] - Ree).real # conductance
        G2 = np.trace(s[0][1]@dagger(s[0][1])).real # total e-e transmission
        return (G1+G2)/2.



def didv_kmap(self,kpath=None,energies=None,
           write=True,**kwargs):
    """Compute the momentum-resolved dIdV"""
    def fun(k,e):
        if self.dimensionality==2: # 2D heterostructure
            HT1 = self.generate(k) # generate heterostructure
            return HT1.didv(energy=e,**kwargs)
        else: raise # not implemented
    if kpath is None: kpath = np.linspace(0.,1.,40)
    if energies is None: energies = np.linspace(-1.0,1.,40)
    from ..parallel import pcall
    kout,eout,dout = [],[],[]
    ds = pcall(lambda k: [fun(k,e) for e in energies],kpath) # call in parallel
    for (k,d) in zip(kpath,ds): # loop over kpoints
        kout = np.concatenate([kout,energies*0.+k]) # store kpoint
        eout = np.concatenate([eout,energies]) # store energies
        dout = np.concatenate([dout,d]) # store DOS
    if write:
        np.savetxt("DIDV_MAP.OUT",np.array([kout,eout,dout]).T)
    return (kout,eout,dout)











import numpy as np
from .. import algebra


def berry_green_generator(f,k=[0.,0.,0.],dk=0.05,operator=None,
              full=False):
  """Function that returns the energy resolved Berry curvature"""
  k = np.array(k) # set as array
  dx = np.array([dk,0.,0.])
  dy = np.array([0.,dk,0.])
  def fint(e): # function to integrate
#    g = f(e=e,k=k) # compute at this k and this energy
    gxp = f(e=e,k=k+dx) # compute at this k and this energy
    gxm = f(e=e,k=k-dx) # compute at this k and this energy
    gyp = f(e=e,k=k+dy) # compute at this k and this energy
    gym = f(e=e,k=k-dy) # compute at this k and this energy
    g = (gxp + gyp + gxm + gym)/4. # average Green function
    # Now apply the formula
    gI = algebra.inv(g) # inverse
    # the derivative of g^-1 is -g^-1*g'*g^-1
    omega = -(gxp-gxm)@gI@(gyp-gym) + (gyp-gym)@gI@(gxp-gxm)
#    omega = ((gxp-gxm)@(gyp-gym) - (gyp-gym)@(gxp-gxm))@gI
#    omega = g*((gxp.I-gxm.I)*(gyp-gym) -(gyp.I-gym.I)*(gxp-gxm))
#    omega += -g*(gyp.I-gym.I)*(gxp-gxm)
    if operator is not None: omega = operator(omega,k=k)
    if full: return omega/(4.*dk*dk*2.*np.pi) # return the full matrix
    else: return np.trace(omega)/(4.*dk*dk*2.*np.pi) # return contribution
  return fint # return the function



def berry_green(f,emin=-10.0,k=[0.,0.,0.],ne=100,dk=0.0001,operator=None):
  """Return the Berry curvature using Green functions"""
  import scipy.integrate as integrate
  fint = berry_green_generator(f,k=k,dk=dk,operator=operator)
  es = np.linspace(emin,0.,ne) # energies used for the integration
  ### The original function is defined in the coplex plane,
  # we will do a change of variables of the form z = re^(iphi) - r0
  # so that dz = re^(iphi) i dphi
  def fint2(x):
    """Function to integrate using a complex contour, from 0 to 1"""
    z0 = emin*np.exp(-1j*x*np.pi)/2.
    z = z0 + emin/2.
    return -(fint(z)*z0).imag*np.pi # integral after the change of variables
  return integrate.quad(fint2,0.0,1.0,limit=60,epsabs=0.1,epsrel=0.1)[0]
#  return integrate.quad(fint,emin,0.0,limit=60,epsabs=0.01,epsrel=0.01)[0]
#  return np.sum([fint(e) for e in es]) # return




def berry_operator(h,delta=1e-1,**kwargs):
    """Return ap operator that computes the Berry curvature for a certain
    wavefunction"""
    if h.dimensionality!=2: raise
    h = h.copy()
    hk = h.get_hk_gen() # get generator
    gk = h.get_gk_gen(delta=delta) # get generator
    if not h.is_sparse: # dense Hamiltonians
        def bk(k): return berry_green_generator(gk,k=k,full=True,**kwargs)
        def outf(w,k=[0.,0.,0.]):
            m = hk(k) # bloch Hamiltonian
            e = algebra.braket_wAw(w,m) # energy
            o = bk(k)(e)@(delta*w) # Berry curvature
            return o # return a vector
        return outf
    else:
        def gI(**kwargs): # workaround for inverse of Green's function
            return gk(**kwargs,inv=True) # make the inverse
        def bk(k): return berry_green_generator_sparse(gk,k=k,gI=gI,**kwargs)
        def outf(w,k=[0.,0.,0.]):
            m = hk(k) # bloch Hamiltonian
            e = algebra.dot(w,m*w) # energy
            w2 = bk(k)(e)*(delta*w) # Operator times WF
            # to think, perhaps this should be <v|B|v>|v>
            # currently it is B|v>
            # (this does not change the expectation value though)
            return w2 # return a vector
        return outf




def berry_green_generator_sparse(f,k=[0.,0.,0.],
        dk=0.05,gI=None):
  """Function that returns the energy resolved Berry curvature"""
  k = np.array(k) # set as array
  dx = np.array([dk,0.,0.])
  dy = np.array([0.,dk,0.])
  def fint(e): # function to return
    gxp = f(e=e,k=k+dx) # compute at this k and this energy
    gxm = f(e=e,k=k-dx) # compute at this k and this energy
    gyp = f(e=e,k=k+dy) # compute at this k and this energy
    gym = f(e=e,k=k-dy) # compute at this k and this energy
    # Now apply the formula
    gI0 = gI(e=e,k=k) # inverse of the Green function
    # the derivative of g^-1 is -g^-1*g'*g^-1
    omega = -(gxp-gxm)@gI0@(gyp-gym) + (gyp-gym)@gI0@(gxp-gxm)
    return omega/(4.*dk*dk*2.*np.pi) # return the full matrix
  return fint # return the function





from ..hamiltonians import Hamiltonian

# TODO
# - redefine get_vev
# - redefine topological invariants
# - redefine get_mean_field_hamiltonian

# quite some functions can be reused from the Green's function formalism

# these implementations still have to be tested


# import the local bandstructure method for non Hermitian
from .bandstructure import get_bands_nd as get_bands
from .dos import get_dos
from .topology import get_berry_curvature




import numpy as np

from .algebra import isnumber,issparse
from . import algebra

is_number = isnumber

class MultiHopping():
    """Class for a multihopping"""
    def __init__(self,a):
        if type(a)==dict: # dictionary type
            self.dict = a # dictionary
        elif type(a)==MultiHopping: # multihopping type
            self.dict = a.dict # dictionary
        elif type(a)==np.ndarray or issparse(a) or type(a)==np.matrix:
            dd = dict() ; dd[(0,0,0)] = a
            self.dict = dd
        else: raise
    def __add__(self,a):
        if type(a)!=MultiHopping: return NotImplemented
        out = add_hopping_dict(self.dict,a.dict)
        out = MultiHopping(out) # create a new object
        return out
    def __mul__(self,a):
        if type(a)==MultiHopping:
            out = multiply_hopping_dict(self.dict,a.dict)
            out = MultiHopping(out) # create a new object
            return out
        elif isnumber(a):
            out = dict()
            for key in self.dict:
                out[key] = a*self.dict[key]
            out = MultiHopping(out) # create a new object
            return out
        else: return NotImplemented
    def __rmul__(self,a): 
        if isnumber(a): return self*a
        else: return NotImplemented
    def __neg__(self): return (-1)*self
    def __sub__(self,a): return self + (-a)
    def get_dict(self):
        return self.dict # dictionary
    def dot(self,a):
        """Compute the dot product"""
        a = MultiHopping(a) # as Multihopping
        return dot_hopping_dict(self.dict,a.dict) # dictionary
    def norm(self):
        """Norm of the MultiHopping"""
        return np.sqrt((self.dot(self)).real)
    def is_hermitian(self):
        dd = self - self.get_dagger()
        if dd.norm()>1e-7: return False
        else: return True
    def get_dagger(self):
        out = dict()
        for key in self.dict:
            key2 = tuple(-np.array(key,dtype=int))
            out[key] = np.conjugate(self.dict[key2].T)
        out = MultiHopping(out) # create a new object
        return out
    def copy(self):
        from copy import deepcopy
        return deepcopy(self)
    def is_zero(self):
        for key in self.dict:
            m = self.dict[key]
            if not algebra.is_zero(m): return False
        return True






def add_hopping_dict(hop1i,hop2i):
    """Multiply two hopping dictionaries"""
    # convert all the keys to tuples
    hop1,hop2 = dict(),dict()
    for key in hop1i: hop1[tuple(key)] = hop1i[key]
    for key in hop2i: hop2[tuple(key)] = hop2i[key]
    # now collect all the matrices
    out = dict() # create dictionary
    keys = [key1 for key1 in hop1]
    for key2 in hop2:
        if key2 not in keys: keys.append(key2) # store
    for key in keys:
        m = 0 # initialize
        if key in hop1: 
            m = (m + hop1[key]).copy()
        if key in hop2: 
            m = (m + hop2[key]).copy()
        out[key] = m.copy() # store
    return out


def dot_hopping_dict(hop1,hop2):
    """Multiply two hopping dictionaries"""
    out = dict() # create dictionary
    out = 0. # initialize
    keys = [key1 for key1 in hop1]
    for key2 in hop2:
        if key2 not in keys: keys.append(key2) # store
    for key in keys:
        m1,m2 = 0.,0.
        if key in hop1: 
            m1 = np.array(algebra.todense(hop1[key]))
        if key in hop2: 
            m2 = np.array(algebra.todense(hop2[key]))
        m = np.sum(np.conjugate(m1)*m2)
        out = out + m # add this contribution
    return out





def multiply_hopping_dict(hop1,hop2):
    """Multiply two hopping dictionaries"""
    out = dict() # create dictionary
    for key1 in hop1:
        for key2 in hop2:
            key = tuple(np.array(key1) + np.array(key2))
            m = hop1[key1]@hop2[key2] # multiply
            if np.max(np.abs(m))>1e-6: # if non-zero
                if key in out: out[key] = out[key] + m # add
                else: out[key] = m # store
    return out # return output




def direct_sum_hopping_dict(hop1,hop2):
    """Perform the direct sum of two hopping dictionaries"""
    from .algebra import direct_sum
    out = dict() # create dictionary
    keys = [key1 for key1 in hop1]
    keys2 = [key1 for key1 in hop2]
    zero1 = 0.*hop1[keys[0]] # initialize as zero
    zero2 = 0.*hop2[keys2[0]] # initialize as zero
    for key2 in hop2:
        if key2 not in keys: keys.append(key2) # store
    for key in keys:
        if key in hop1: m1 = hop1[key]
        else: m1 = zero1
        if key in hop2: m2 = hop2[key]
        else: m2 = zero2
        out[key] = direct_sum([m1,m2]) # store
    return out




# redefine topological invariants

def get_berry_curvature(self,**kwargs):
    """Compute Berry curvature"""
    from ..topology import get_berry_curvature_master as BC
    return BC(self,mode="Green",**kwargs)



from __future__ import print_function
from scipy.sparse import csc_matrix,bmat
import numpy as np
from .multihopping import MultiHopping
from. import superconductivity
#from .scftypes import selfconsistency

dup = csc_matrix(([1.0],[[0],[0]]),shape=(2,2),dtype=np.complex_)
ddn = csc_matrix(([1.0],[[1],[1]]),shape=(2,2),dtype=np.complex_)
sp = csc_matrix(([1.0],[[1],[0]]),shape=(2,2),dtype=np.complex_)
sm = csc_matrix(([1.0],[[0],[1]]),shape=(2,2),dtype=np.complex_)
def zero(d): return csc_matrix(([],[[],[]]),shape=(d,d),dtype=np.complex_)


def element(i,n,p,d=2,j=None):
  if j is None: j=i
  o = csc_matrix(([1.0],[[d*i+p[0]],[d*j+p[1]]]),
                   shape=(n*d,n*d),dtype=np.complex_)  
  return o
  o = csc_matrix(([1.0],[[p[0]],[p[1]]]),shape=(d,d),dtype=np.complex_)
  m = [[None for i1 in range(n)] for i2 in range(n)]
  for i1 in range(n): m[i1][i1] = zero(d)
  m[i][j] = o.copy()
  return bmat(m) # return matrix



class interaction(): 
  def __init__(self):
    self.contribution = "AB" # assume that both operators contribute
    self.dir = [0,0,0] # unit cell to which to hop
    self.dhop = [0,0,0] # matrix that is affected

def hubbard_density(i,n,g=1.0):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  v.a = element(i,n,[0,0]) # value of the coupling
  v.b = element(i,n,[1,1]) # value of the coupling
  v.dir = [0,0,0] # direction of the interaction 
  v.g = g
  v.i = i
  v.j = i
  return v



def spinless_hubbard_density(i,n,g=1.0):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  v.a = element(i,n,[0,0],d=1) # value of the coupling
  v.b = element(i,n,[0,0],d=1) # value of the coupling
  v.dir = [0,0,0] # direction of the interaction 
  v.g = g
  v.i = i
  v.j = i
  return v







def hubbard_exchange(i,n,g=1.0):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  v.a = element(i,n,[0,1]) # value of the coupling
  v.b = element(i,n,[1,0]) # value of the coupling
  v.dir = [0,0,0] # direction of the interaction 
  v.g = -g # minus from fermion
  v.i = i
  v.j = i
  return v




def hubbard_pairing_ud(i,n,g=1.0):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  v.a = element(i,n,[0,2],d=4) # cc
  v.b = element(i,n,[2,0],d=4) # cdcd
  v.dir = [0,0,0] # direction of the interaction 
  v.g = g
  v.i = i
  v.j = i
  return v


def hubbard_pairing_du(i,n,g=1.0):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  v.a = element(i,n,[1,3],d=4) # cc
  v.b = element(i,n,[3,1],d=4) # cdcd
  v.dir = [0,0,0] # direction of the interaction 
  v.g = g
  v.i = i
  v.j = i
  return v




def v_pairing_uu(i,j,n,g=1.0,d=[0,0,0],channel="ee"):
  """Return pair of operators for a Hubbard mean field"""
  v = interaction()
  if channel=="ee": # ee channel
    v.a = element(i,n,[0,3],d=4,j=j) # cc
    v.b = element(j,n,[3,0],d=4,j=i) # cdcd
  elif channel=="hh":
    v.a = element(i,n,[3,0],d=4,j=j) # cc
    v.b = element(j,n,[0,3],d=4,j=i) # cdcd
  else: raise
  v.dir = d # direction of the interaction 
  v.g = g
  v.contribution = "A"
  v.i = i
  v.j = j
  return v


def v_pairing_dd(i,j,n,g=1.0,d=[0,0,0],channel="ee"):
  """Return pair of operators for a V mean field"""
  v = interaction()
  if channel=="ee": # ee channel
    v.a = element(i,n,[1,2],d=4,j=j) # cc
    v.b = element(j,n,[2,1],d=4,j=i) # cdcd
  elif channel=="hh":
    v.a = element(i,n,[2,1],d=4,j=j) # cc
    v.b = element(j,n,[1,2],d=4,j=i) # cdcd
  else: raise
  v.dir = d # direction of the interaction 
  v.g = g
  v.contribution = "A"
  v.i = i
  v.j = j
  return v


def v_pairing_du(i,j,n,g=1.0,d=[0,0,0]):
  """Return pair of operators for a V mean field"""
  v = interaction()
  v.a = element(i,n,[1,3],d=4,j=j) # cc
  v.b = element(j,n,[3,1],d=4,j=i) # cdcd
  v.dir = d # direction of the interaction 
  v.g = g
  v.i = i
  v.j = j
  return v




def v_ij(i,j,n,g=1.0,d=[0,0,0],spini=0,spinj=0):
  """Return pair of operators for a V mean field"""
  v = interaction()
  v.a = element(i,n,[spini,spini],d=2,j=j) # cc
  v.b = element(j,n,[spinj,spinj],d=2,j=i) # cdc
  v.dir = d # direction of the interaction 
  v.g = -g # this minus comes from commutation relations
  v.contribution = "A"
  v.i = i
  v.j = j
  return v


def v_ij_spinless(i,j,n,g=1.0,d=[0,0,0]):
  """Return pair of operators for a V mean field"""
  v = interaction()
  v.a = csc_matrix(([1.0],[[i],[j]]),shape=(n,n),dtype=np.complex_) # cc
  v.b = csc_matrix(([1.0],[[j],[i]]),shape=(n,n),dtype=np.complex_) # cdc
  v.dir = d # direction of the interaction 
  v.dhop = d # direction of the interaction 
  v.g = -g # this minus comes from commutation relations
  v.contribution = "A"
  v.i = i
  v.j = j
  return v


def v_ij_density_spinless(i,j,n,g=1.0,d=[0,0,0],contribution="AB"):
  """Return pair of operators for a V mean field"""
  v = interaction()
  v.a = csc_matrix(([1.0],[[i],[i]]),shape=(n,n),dtype=np.complex_) # cc
  v.b = csc_matrix(([1.0],[[j],[j]]),shape=(n,n),dtype=np.complex_) # cdc
  v.dir = d # direction of the neighbor
  v.dhop = [0,0,0] # hopping that is affected
  v.g = g 
  v.contribution = contribution
  v.i = i
  v.j = j
  return v


def v_ij_fast_coulomb(i,jvs,n,vcut=1e-3):
  """Return the interaction part for the fast Coulomb trick"""
  v = interaction()
  v.a = csc_matrix(([1.0],[[i],[i]]),shape=(n,n),dtype=np.complex_) # cc
  jj = range(n) # indexes
  if len(jvs)!=n: raise # something wrong
  v.b = csc_matrix((jvs,[jj,jj]),shape=(n,n),dtype=np.complex_) # cdc
  v.b.eliminate_zeros()
  v.dir = [0,0,0] # direction of the neighbor
  v.dhop = [0,0,0] # hopping that is affected
  v.g = 1.0 # default value
  v.contribution = "A"
  return v



def v_ij_fast_coulomb_spinful(i,jvs,n,channel="up"):
  """Return the interaction part for the fast Coulomb trick"""
  jvs2 = np.zeros(2*n) # initialize
  if channel=="up":
      for jj in range(n): jvs2[2*jj] = jvs[jj]
      ii = 2*i+1
  elif channel=="down":
      for jj in range(n): jvs2[2*jj+1] = jvs[jj]
      ii = 2*i
  else: raise
  return v_ij_fast_coulomb(ii,jvs2,2*n)



symmetry_breaking = ["magnetic","antiferro","ferroX","ferroY",
        "ferroZ","Charge density wave","Haldane",
        "kanemele","rashba","s-wave superconductivity"]

spinful_guesses = ["Fully random","dimerization"]

spinful_guesses += symmetry_breaking


def guess(h,mode="ferro",fun=1e-1):
  """Return a mean field matrix guess given a certain Hamiltonian"""
  h0 = h.copy() # copy Hamiltonian
  h0 = h0.get_multicell() # multicell
#  h0.intra *= 0. # initialize
  h0.clean() # clean the Hamiltonian
  if mode=="ferro":
      if h.has_spin: h0.add_zeeman(fun)
  elif mode=="magnetic":
      if h.has_spin: h0.add_zeeman(lambda x: np.random.random(3)*fun)
  elif mode=="ferroX":
      if h.has_spin: h0.add_zeeman([fun,0.,0.])
  elif mode=="ferroY":
      if h.has_spin: h0.add_zeeman([0.,fun,0.])
  elif mode=="ferroZ":
      if h.has_spin: h0.add_zeeman([0.,0.,fun])
  elif mode=="randomXY" or mode=="XY":
      def f(r):
          m = [np.random.random()-0.5,np.random.random()-0.5,0.]
          m = np.array(m)
          return m/np.sqrt(m.dot(m))
      if h.has_spin: h0.add_zeeman(f)
      return h0.get_hopping_dict()
  elif mode=="random":
      dd = h.get_dict()
      for key in dd:
          n = dd[key].shape[0]
          dd[key] = np.random.random((n,n))-.5 + 1j*(np.random.random((n,n))-.5)
      dd = MultiHopping(dd)
      dd = dd + dd.get_dagger()
      return dd.get_dict()
  elif mode=="dimerization":
      return guess(h,mode="random",fun=fun)
  elif mode=="kekule":
      h0.turn_multicell()
      h0.add_kekule(fun) # Haldane coupling
      return h0.get_hopping_dict()
  elif mode=="Haldane":
      h0.add_haldane(fun) # Haldane coupling
      return h0.get_hopping_dict()
  elif mode=="rashba":
      if h.has_spin: h0.add_rashba(fun) # Haldane coupling
      return h0.get_hopping_dict()
  elif mode=="kanemele":
      if h.has_spin: h0.add_kane_mele(fun) # Haldane coupling
      return h0.get_hopping_dict()
  elif mode in ["antihaldane","valley"]:
      h = h.copy() ; h.clean() ; h.add_antihaldane(fun) # Haldane coupling
      return h.get_hopping_dict()
  elif mode=="Fully random": return None
  elif mode in ["CDW","Charge density wave"]:
      if h.geometry.has_sublattice:
        h0.add_onsite(h.geometry.sublattice)
      else: return 0.0 #guess(h,mode="random",fun=0.0)
  elif mode=="potential":
      h0.add_onsite(fun)
  elif mode=="antiferro":
      if h.has_spin: h0.add_antiferromagnetism(fun)
  elif mode=="imbalance":
      h0.add_sublattice_imbalance(fun)
  elif mode in ["swave","s-wave superconductivity"]:
      if h.has_eh: h0.add_swave(fun)
  elif mode=="pwave":
    for t in h0.hopping: t.m *= 0. # clean
    h0.add_pwave(fun)
    hop = dict()
    hop[(0,0,0)] = h0.intra
    for t in h0.hopping: hop[tuple(t.dir)] = t.m
    return hop
  else:
      print("Unrecognized initialization")
      raise
  return h0.intra # return matrix

from .algebra import braket_wAw
#from numba import jit

#@jit
def expectation_value(wfs,A,phis):
  """Return the expectation value of a set of wavevectors"""
  out = 0.0j
  for (p,w) in zip(phis,wfs):
    out += braket_wAw(w,A)*p
#    w = np.matrix(w) # convert to matrix
#    out += ((w.T).H*A*w.T)[0,0]*p # expectation value
  return np.conjugate(out) # return value




def enforce_pwave(mf):
  """Enforce pwave symmetry in a mean field Hamiltonian"""
  for key in mf: mf[key] = np.matrix(mf[key]) # dense matrix
  for key in mf: 
#    print(mf[key],type(mf[key]))
    n = mf[key].shape[0]//4 # number of sites 
    dm = tuple([-di for di in key]) # the opposite matrix
    for i in range(n): # loop over positions
      for j in range(n): # loop over positions
        for (ii,jj) in [(1,2),(0,3),(2,1),(3,0)]:
          mf[key][4*i+ii,4*j+jj] = (mf[key][4*i+ii,4*j+jj] - mf[dm][4*j+ii,4*i+jj])/2.
  return mf


def enforce_eh(h,mf):
  """Enforce eh symmetry in a mean field Hamiltonian"""
  from .superconductivity import eh_operator
  eh = eh_operator(h.intra) # get the function
  for key in mf: mf[key] = np.matrix(mf[key]) # dense matrix
  mfout = dict()
  for key in mf:
    mkey = (-key[0],-key[1],-key[2]) 
    mfout[key] = (mf[key] - eh(mf[mkey].H))/2.
  return mfout



def broyden_solver(scf):
    """Broyden solver for selfconsistency"""
    scf.mixing = 1.0 # perfect mixing
    scf.iterate() # do one iteration
    x0 = scf.cij # get the array with expectation values
    def fun(x): # function to solve
        scf.cij = x # impose those expectation values
        scf.cij2v() # update mean field
        scf.iterate() # perform an iteration
        return scf.cij - x # difference with respect to input
    from scipy.optimize import broyden1,broyden2,anderson
#    x = anderson(fun,x0)
    x = broyden1(fun,x0,f_tol=1e-7) # broyden solver
    scf.cij = x # impose those expectation values
    scf.cij2v() # update mean field
    scf.iterate() # perform an iteration
    return scf # return scf




def coulomb_interaction(g,vc=1.0,vcut=1e-4,vfun=None,has_spin=False,**kwargs):
    """Return a list with the Coulomb interaction terms"""
    interactions = [] # empty list
    nat = len(g.r) # number of atoms
    lat = np.sqrt(g.a1.dot(g.a1)) # size of the unit cell
    rcut = 4.0
    g.ncells = max([int(2*rcut/lat),1]) # number of unit cells to consider
    ri = g.r # positions
    if vfun is None: # no function provided
        def vfun(dr):
            if dr<1e-4: return 0.0
            if dr>rcut: return 0.0
            return vc/dr*np.exp(-dr/rcut)
    for d in g.neighbor_directions(): # loop
      rj = np.array(g.replicas(d)) # positions
      for i in range(nat): # loop over atoms
        for j in range(nat): # loop over atoms
          dx = rj[j,0] - ri[i,0]
          dy = rj[j,1] - ri[i,1]
          dz = rj[j,2] - ri[i,2]
          dr = dx*dx + dy*dy + dz*dz
          dr = np.sqrt(dr) # square root
          v = vfun(dr) # interaction
          if v>1e-3: # sizeble interaction
            if has_spin:
              interactions.append(v_ij_density_spinless(2*i,2*j+1,2*nat,
              g=v,d=[0,0,0]))
            else:
              interactions.append(v_ij_density_spinless(i,j,nat,
              g=v,d=[0,0,0]))
    return interactions


def coulomb_interaction_spinless(g,**kwargs):
    return coulomb_interaction(g,has_spin=False,**kwargs)


def coulomb_interaction_spinful(g,**kwargs):
    return coulomb_interaction(g,has_spin=True,**kwargs)



def fast_coulomb_interaction(g,vc=1.0,vcut=1e-4,vfun=None,has_spin=False,**kwargs):
    """Return a list with the Coulomb interaction terms, summed over sites"""
    interactions = [] # empty list
    nat = len(g.r) # number of atoms
    lat = np.sqrt(g.a1.dot(g.a1)) # size of the unit cell
    rcut = 4.0
    g.ncells = max([int(2*rcut/lat),1]) # number of unit cells to consider
    ri = g.r # positions
    if vfun is None: # no function provided
        def vfun(dr):
            if dr<1e-4: return 0.0
            if dr>rcut: return 0.0
            return vc/dr*np.exp(-dr/rcut)
    for i in range(nat): # loop over atoms
      vjs = np.zeros(nat) # initialize
      for d in g.neighbor_directions(): # loop
        rj = np.array(g.replicas(d)) # positions
        for j in range(nat): # loop over atoms
          dx = rj[j,0] - ri[i,0]
          dy = rj[j,1] - ri[i,1]
          dz = rj[j,2] - ri[i,2]
          dr = dx*dx + dy*dy + dz*dz
          dr = np.sqrt(dr) # square root
          vt = vfun(dr) # interaction
          vjs[j] += vt # add contribution
      vjs[vjs<1e-4] = 0.0 # discard near zeros
      if np.sum(vjs)>1e-3: # sizable interaction
        print("Total Coulomb term",np.sum(vjs))
        if has_spin:
          interactions.append(
                  v_ij_fast_coulomb_spinful(i,vjs,nat,channel="up")
                  )
          interactions.append(
                  v_ij_fast_coulomb_spinful(i,vjs,nat,channel="down")
                  )
        else: raise
    return interactions


def identify_symmetry_breaking(h0,h,as_string=False,tol=1e-3):
    """Given two Hamiltonians, identify what is the symmetry
    breaking between them"""
    dt0 = h0.get_multihopping() # first multihopping
    dt = h.get_multihopping() # second multihopping
    dd = dt0 - dt # difference between Hamiltonians
    out = [] # empty list
    if dd.norm()>tol: # non-zero
        for s in symmetry_breaking: # loop over contributions
            try:
                d0 = MultiHopping(guess(h,s,fun=1.0)) # get this type
                proj = dd.dot(d0) # compute the projection
                if np.abs(proj)>tol: out.append(s)
            except: pass
        h1 = h0.copy() # copy
        h1.set_multihopping(dd) # set the difference
        out += superconductivity.identify_superconductivity(h1,tol=tol) # check SC
    else: out = ["No symmetry breaking"]
    if len(out)==0: out = ["Unidentified symmetry breaking"]
    if as_string: # return as a string
        out2 = ""
        for o in out: out2 += o + ", "
        return out2
    return out


def order_parameter(self,name):
    mf = self.hamiltonian - self.hamiltonian0 # mean field
    if name=="even_SC":
        from .sctk.orderparameter import singlet
        return singlet(mf)
    elif name=="odd_SC":
        from .sctk.orderparameter import triplet
        return triplet(mf)
    else: raise




from .selfconsistency import densitydensity

hubbardscf = densitydensity.hubbard
Vinteraction = densitydensity.Vinteraction

from .selfconsistency.potentials import keldysh



# library to verify the type of Hamiltonian
import numpy as np

def check_mode(h,n):
    if n=="spinless_nambu":
        if (not h.has_spin) and h.has_eh: return True
        else: return False
    elif n=="spinful_nambu":
        if h.has_spin and h.has_eh: return True
        else: return False
    elif n=="spinful":
        if h.has_spin and (not h.has_eh): return True
        else: return False
    elif n=="spinless":
        if (not h.has_spin) and (not h.has_eh): return True
        else: return False
    else: raise



def reduce_hamiltonian(self):
    """Try to reduce the dimensionality of the Hamiltonian"""
    if self.check_mode("spinless"): return self # nothing to do
    elif self.check_mode("spinful"): # spinful Hamiltonian
        h = self.copy() # copy Hamiltonian
        h.remove_spin() # remove the spin degree of freedom
        h0 = h.copy() # copy Hamiltonian
        h.turn_spinful() # add spin degree of freedom
        if self.same_hamiltonian(h): 
#            print("Hamiltonian has become spinless")
            return h0 # return the spinless one
        else: return self # return the spinful one
        return self # nothing to do
    else: return self # nothing to do


def same_hamiltonian(self,h,ntries=10):
    """Check if two hamiltonians are the same"""
    if h.dimensionality==0: ntries = 1
    hk1 = self.get_hk_gen()
    hk2 = h.get_hk_gen()
    for i in range(ntries):
      k = np.random.random(3)
      m = hk1(k) - hk2(k)
      if np.max(np.abs(m))>0.000001: return False
    return True


def turn_spinful(self,enforce_tr=False):
    """Turn the hamiltonian spinful"""
    if self.has_spin: return # already spinful
    if self.is_sparse: # sparse Hamiltonian
      self.turn_dense() # dense Hamiltonian
      self.turn_spinful(enforce_tr=enforce_tr) # spinful
      self.turn_sparse()
    else: # dense Hamiltonian
      from ..increase_hilbert import spinful
      def fun(m):
          if enforce_tr: return spinful(m,np.conjugate(m))
          else: return spinful(m)
      self.modify_hamiltonian_matrices(fun) # modify the matrices
      self.has_spin = True # set spinful


def make_compatible(h1,h2):
    """Make a Hamiltonian h1 compatible with hamiltonian h2"""
    ho = h1.copy() # copy Hamiltonian
    if h1.has_spin: ho.turn_spinful() # make spinful
    if h2.has_eh: ho.turn_nambu() # add e-h symmetry
    return ho





import numpy as np
from ..spin import sx,sy,sz,bmat
from ..superconductivity import build_eh


def get_si(h,i=1):
    """Return a certain Pauli matrix for the full Hamiltonian"""
    if not h.has_spin: return None # no spin
    if i==1: si = sx # sx matrix
    elif i==2: si = sy # sy matrix
    elif i==3: si = sz # sz matrix
    else: raise # unknown pauli matrix
    if h.has_eh: ndim = h.intra.shape[0]//4 # half the dimension
    else: ndim = h.intra.shape[0]//2 # dimension
    if h.has_spin: # spinful system
      op = [[None for i in range(ndim)] for j in range(ndim)] # initialize
      for i in range(ndim): op[i][i] = si # store matrix
      op = bmat(op) # create matrix
    if h.has_eh: op = build_eh(op,is_sparse=True) # add electron and hole parts
    return op

# define the functions for the three spin components
get_sx = lambda h: get_si(h,i=1) # sx matrix
get_sy = lambda h: get_si(h,i=2) # sy matrix
get_sz = lambda h: get_si(h,i=3) # sz matrix








import numpy as np
from .. import algebra
from ..green import green_renormalization
from .. import green
from copy import deepcopy
from ..htk.mode import make_compatible

delta_smatrix = 1e-12
dagger = algebra.dagger
gfmode = "adaptive"

# library to perform transport calculations using a local probe


class LocalProbe():
    def __init__(self,h,lead=None,delta=1e-5,i=0,T=1.0,**kwargs):
        self.H = h.copy() # store Hamiltonian
        self.H.turn_dense() # dense Hamiltonian
        self.has_eh = self.H.has_eh # electron-hole
        self.delta = delta
        self.mode = "bulk"
        self.reuse_gf = False # reuse the Green's function
        self.gf = None
        self.bulk_delta = delta
        self.frozen_lead = True
        self.i = i # this site
        if lead is None:
            from ..geometry import chain 
            lead = chain().get_hamiltonian(has_spin=False) # create a chain
        lead = make_compatible(lead,self.H) # make them compatible
        lead = lead.get_no_multicell() # no multicell
        self.lead = lead.copy() # store
        self.get_eh_sector = self.lead.get_eh_sector # if it has electron-hole
        self.T = T # transparency
    def get_selfenergy(self,energy,lead=0,**kwargs):
        """Return the selfenergies"""
        if lead==0: # use the probe
            return lead_selfenergy(self,energy=energy,**kwargs)
        elif lead==1: # use the system
            g = generate_gf(self,energy=energy,
                               **kwargs) # generate the Green's function
            return local_selfenergy(self.H,g,i=self.i,
                                energy=energy,**kwargs)
        else: raise
    def get_central_gmatrix(self,**kwargs):
        return get_central_gmatrix(self,**kwargs)
    def get_reflection_normal_lead(self,s):
        return get_reflection_normal_lead(self,s)
    def didv(self,T=None,**kwargs):
        from .didv import didv
        return didv(self,**kwargs)
    def copy(self): return deepcopy(self)
    def set_coupling(self,c):
        self.T = c # set the coupling
    def remove_pairing(self):
        self.H.remove_pairing()
        self.lead.remove_pairing()
    def get_kappa(self,T=None,**kwargs):
        from .kappa import get_kappa_ratio
        if T is None: T = self.T 
        return get_kappa_ratio(self,T=T,**kwargs)




def generate_gf(self,energy=0.0,**kwargs):
    """Generate the specific Green's function"""
    mode = self.mode 
    # just a trick to reuse the GF if needed
    if self.reuse_gf and self.gf is not None: return self.gf 
    else:
        gf = self.H.get_gf(energy=energy,delta=self.bulk_delta,
                             mode=gfmode,
                             gtype=mode)
        if self.reuse_gf: self.gf = gf # overwrite
        return gf


def lead_selfenergy(self,energy=0.0,**kwargs):
     """Return the selfenergy of the lead"""
     if self.frozen_lead: energy = 0.0 # set as zero energy
     delta = self.delta
     intra = self.lead.intra
     inter = dagger(self.lead.inter)
     cou = inter
     ggg,g = green_renormalization(intra,inter,
                                     energy=energy,
                                     delta=delta)
     sigma = cou@g@dagger(cou) # selfenergy
     return sigma

from ..htk.extract import local_hamiltonian

def local_selfenergy(h,g,energy=0.0,i=0,delta=1e-5,**kwargs):
    """Given a certain Hamiltonian and Green's function, extract
    the local selfenergy"""
    M = get_intra(h) # get intracell matrix
    gi = local_hamiltonian(h,g,i=i) # local Green's function
    oi = local_hamiltonian(h,M,i=i) # local Hamiltonian
    iden = np.identity(gi.shape[0],dtype=np.complex_)
    out = algebra.inv(gi) - (energy+1j*delta)*iden + oi # local selfenergy
    return -out



def get_central_gmatrix(P,selfl=None,selfr=None,energy=0.0):
    """Return the central Green's function"""
    delta = P.delta # imaginary part
    if selfl is None: selfl = P.get_selfenergy(lead=0,energy=energy)
    if selfr is None: selfr = P.get_selfenergy(lead=1,energy=energy)
    if delta>delta_smatrix: delta = delta_smatrix # small delta is critical!
    iden = np.identity(selfl.shape[0],dtype=complex)*(energy +1j*delta)
    if P.frozen_lead:
        idenl = np.identity(selfl.shape[0],dtype=complex)*1j*delta
    else: idenl = iden
    hlist = [[None for i in range(2)] for j in range(2)] # list of matrices
    M = get_intra(P.H) # intracell matrix
    oi = local_hamiltonian(P.H,M,i=P.i) # local Hamiltonian
    # set up the different elements
    # first the intra terms
    hlist[0][0] = idenl - P.lead.intra - selfl
    hlist[1][1] = iden - oi - selfr
    # now the inter cell
    hlist[0][1] = -P.lead.inter*P.T # coupling times transparency
    hlist[1][0] = dagger(hlist[0][1]) # Hermitian conjugate
    return hlist


def get_reflection_normal_lead(P,s):
    return s[0][0]




def get_intra(H):
    """Function to extract the intra-matrix from an object
    depending on its type"""
    from ..hamiltonians import Hamiltonian
    from ..embedding import Embedding
    if type(H)==Hamiltonian: return H.intra
    elif type(H)==Embedding: return H.m
    else: raise



def Hamiltonian_didv(self,**kwargs):
   """Wrapper to compute the didv for a Hamiltonian"""
   lp = LocalProbe(self,**kwargs) # create the object
   return lp.didv(**kwargs)







# routines to compute the stacking in multilayers
import numpy as np
from numba import jit


def stacking(g):
    """Return the array with the stacking for a gemetry"""
    rs = g.multireplicas(1) # replicated
    return circle_overlap(g.r,rs)


def circle_overlap(rs0,rs1):
    """Return the circle overlap for a set of locations"""
    out = []
    for r in rs0:
        z = r[2] # get the z location 
        dzs = np.abs(rs1[:,2] - z) # distance to the closest layer
        dz = np.min(dzs[dzs>1e-5]) # closest distance
        cls = np.abs(dzs - dz) # this will be zero for the closest layer
        rs = rs1[cls<1e-5] # select atoms of the closest layers
        rs = rs - r # shift by the specific atom
        rs = rs[:,0:2] # retain only xy coordinates
        drs = np.sum(rs*rs,axis=1) # compute distance
        drs = drs[drs<(10.)] # only closest sites
        o = circle_overlap_jit(drs) # compute the overlap
        out.append(o) # store
    return np.array(out) # return the stacking

@jit(nopython=True)
def circle_overlap_jit(drs):
    out = 0. # initialize
    for dr in drs: # loop over distances
        out += np.exp(-dr) # this is just a workaround
    return out




import multicell
import scipy.linalg as lg

def effective2d(hin,k0 = 0.0,ewindow = [-1.0,1.0]):
  """Calculates the effective Hamiltonian for a 2d system"""
  h = multicell.turn_multicell(hin) # multicell form
  hkgen = h.get_hk_gen() # get hamiltonian generator
  (es,ws) = lg.eigh(hkgen(k0)) # initial waves
  ws = np.transpose(ws) # transpose the waves
  wf0 = [] # list to store waves
  for i in range(len(es)): # loop over energies
    if ewindow[0]<ewindow[1]: # check whether in window
      wf0.append(ws[i]) # store this wave
  # now calculate the effective Hamiltonian 
  raise
  for order in orders: # loop over orders
    dh = multicell.derivative(h,k0,order=order) # derivative of the hamiltonian




from __future__ import print_function, division
import numpy as np
from .algebra import isnumber

class Potential():
    def __init__(self,f,g=None):
        if type(f)==Potential: self.f = f.f # store function
        elif callable(f): self.f = f # store function
        elif isnumber(f): self.f = lambda r: f # store function
        else: 
            print("Unrecognized potential",f)
            raise
        self.g = g # geometry
    def __add__(self,a):
        a = Potential(a)
        g = lambda r: self.f(r) + a.f(r)
        return Potential(g,g=self.g)
    def redefine(self,f):
        g = lambda r: f(self.f(r))
        return Potential(g,g=self.g)
    def __mul__(self,a):
        a = Potential(a)
        g = lambda r: self.f(r)*a.f(r)
        return Potential(g,g=self.g)
    def __rmul__(self,a): return self*a
    def __neg__(self): return (-1)*self
    def __sub__(self,a): return self + (-1)*a
    def __rsub__(self,a): return -self + a
    def __radd__(self,a): return self + a
    def __truediv__(self,a): return self*(1./a)
    def __call__(self,r):
        return self.f(r)
    def normalize(self):
        return Potential(enforce_minmax(self,[0.,1.],g=self.g),g=self.g)
    def set_average(self,average):
        out = enforce_average(self,average,g=self.g)
        return Potential(out,g=self.g)



def cnpot(n=4,k=None,v=1.0,angle=0.,r0=np.array([0.,0.,0.])):
  """Returns a function that generates a potential
  with C_n symmetry"""
  if k is None: raise
  if n==0: return lambda r: v
  if n%2==0: f = np.cos # even 
  if n%2==1: f = np.sin # even 
  def fun(r):
    """Function with the potential"""
    r = r - r0 # shift
    x0,y0 = r[0],r[1]
    def Rk(k0,angle):
      x0,y0 = k0[0],k0[1] # components
      x = np.cos(angle)*x0 + np.sin(angle)*y0
      y = np.cos(angle)*y0 - np.sin(angle)*x0
      return np.array([x,y,k0[2]]) # return rotated vector
    acu = 0. # result
    for i in range(n):
      ki = Rk(k,np.pi*2*i/n)
      acu += f(2.*np.pi*ki.dot(r)) 
    return v*acu/n
  return Potential(fun)




def aahf1d(n0=0,beta=1e-7,k=None,b=None,v=1.0,normalize=False):
  """Return the generalized AAHF potential"""
  tau = (1.+np.sqrt(5))/2.
  if b is None: b = 1/tau # default field
  if k is None: k = 3*np.pi*b # default phase
  if beta==0.0: beta=0.000001 # just in case
  def fun(r):
    """Function"""
    ns = r[0] # first coordinate
    ys = np.tanh(beta*(np.cos(2.*np.pi*b*ns+k)-np.cos(np.pi*b)))
    ys /= np.tanh(beta)
    return v*ys
  return Potential(fun) # return function



def commensurate_potential(g,k=1,amplitude=1.0,n=None,
        average=0.0,minmax=None,**kwargs):
    """Return a potential that is commensurate with
    the lattice"""
    if g.dimensionality==2:
      a12 = g.a2.dot(g.a1)/(np.sqrt(g.a1.dot(g.a1))*np.sqrt(g.a1.dot(g.a1)))
      if n is None:
        if 0.49<abs(a12)<0.51: n = 6
        elif abs(a12)<0.01: n = 4
        else: n = 3
      f = cnpot(n=n,k=k*g.b1,**kwargs)
    elif g.dimensionality==1: 
      f = cnpot(n=1,k=k*g.b1,**kwargs)
    else: raise
    f = enforce_amplitude(f,amplitude,g=g) # enforce the amplitude
    f = enforce_average(f,average,g=g) # enforce average
    if minmax is not None: f = enforce_minmax(f,minmax,g=g) # enforce minmax
    f = Potential(f)
    f.g = g
    return f


def impurity(r0,v=0.):
    """Create the potential for an impurity"""
    def f(r):
        dr = r-r0
        if dr.dot(dr)<1e-3: return v
        else: return 0.
    return Potential(f) # return the potential



def fibonacci(n,n0=0):
    """Generate the Fibonacci sequence"""
    a1 = [0] # first
    a2 = [0,1] # first
    while True:
      a3 = a1 + a2
      a1 = a2
      a2 = a3
      if len(a3)>n+n0: break
    out = [a3[i] for i in range(n0,n0+n)] # output
    return np.array(out)




def thue_morse(n,n0=0):
  """Generate the Thue-Morse sequence"""
  out = []
  for i in range(n0,n0+n): # loop over integers
    ib = bin(i).split("b")[1] # in binary
    acu = 0 # start
    for iib in ib: # loop
      if iib=="1": acu += 1
    acu = acu%2 # modulus
    out.append(acu) # output
  return np.array(out)



def tbgAA(g):
    """Return a function that yields +1 for AA and -1 otherwise"""
    from . import geometry
    from .specialhopping import twisted_matrix
    h = g.get_hamiltonian(mgenerator=twisted_matrix(ti=0.3),
            has_spin=False,is_sparse=True)
    h.set_filling(0.5,nk=1)
    ks = [[.5,0.,0.],[0.,.5,0.],[.5,.5,0.]]
    (x,y,d) = h.get_ldos(e=0.0,delta=0.01,ks=ks,mode="arpack",
            write=False,silent=True,nrep=1)
    d = d - np.mean(d) # average in zero
    d = d - np.min(d)
    d = d/np.max(d) # set maximum in 1
    d = 2*(d-0.5) # between -1 and +1
    funr = geometry.get_fractional_function(g,center=True)
    rf = np.array([funr(ri) for ri in g.r])
    fint = interpolate2d(rf[:,0:2],d) # interpolation
    return Potential(lambda ri: fint(funr(ri))[0])



def interpolate2d(r,v):
    """Return a function that does 2d interpolation"""
    from scipy.interpolate import interp2d
    x,y = r[:,0],r[:,1] # data
    grid_x, grid_y = np.mgrid[np.min(x):np.max(x):100j,np.min(y):np.max(y):100j]
    from scipy.interpolate import griddata
    grid_z = griddata(r,v, (grid_x, grid_y), method='nearest')
    f = interp2d(grid_x, grid_y, grid_z, kind='linear')
    return lambda ri: f(ri[0],ri[1])




def enforce_average(f,a,g=None):
    """Normalize the average value of a function for the geometry"""
    if g is None: raise
    m = np.mean([f(ri) for ri in g.r]) # average value
    def fout(r):
        return f(r) + a - m # return this value
    return fout # return new function


def enforce_amplitude(f,a,g=None):
    """Normalize the average value of a function for the geometry"""
    if g is None: raise
    vs = [f(ri) for ri in g.r] 
    minv = np.min(vs)
    maxv = np.max(vs)
    dv = maxv-minv # amplitude
    def fout(r):
        return f(r)*a/dv # return this value
    return fout # return new function



def enforce_minmax(f,a,g=None):
    """Enforce the minimum, without changing the maximum"""
    if g is None: raise
    vs = [f(ri) for ri in g.r] 
    minv = np.min(vs)
    maxv = np.max(vs)
    dv = maxv-minv # amplitude
    fout = lambda r: (f(r)-minv)/dv*a[1] + (1.-(f(r)-minv)/dv)*a[0] # return this value
    return fout # return new function



def array2potential(x,y,v):
    """Given an initial xyz array, return a function
    that interpolates over them"""
    from .interpolation import interpolator2d
    if len(v)!=len(x): raise
    return Potential(interpolator2d(x,y,v))


def object2potential(V,r=None):
    """Transform a generic object into a callable potential"""
    # this should be finished
    return V


from .potentialtk.profiles import radial_decay


def delta(r0,v0=1.0):
    """Delta function potential"""
    def f(r):
        dr = r - r0
        dr2 = dr.dot(dr)
        if dr2<1e-4: return v0
        else: return v0*0.0
    return Potential(f) # return potential



def edge_potential(g):
    """Return potential for the edge sites"""
    cs = g.get_connections() # get the connections
    nn = np.max([len(c) for c in cs]) # maximum number of connections
    v = [len(c)<nn for c in cs] # check if the atom is an edge 
    fout = 0 # output function
    for i in range(len(v)): # loop over locations
        if v[i]: fout = fout + delta(g.r[i],v0=1.0)
    return Potential(fout) # return the potential



def stacking_potential(g,**kwargs):
    """Given a certain geometry, return a function with that gives the
    stacking"""
    # this function should be polished
    from .crystalfield import cf_potential
#    g1 = g.copy() ; g1.r[:,2] = 0.0
    v = cf_potential(g,vc=2.0,mode="stacking",**kwargs)
    v = v - np.min(v) # shift to zero
    v = v/np.max(v) # normalize
    v = v - 0.5 # between -0.5 and 0.5
    from .geometry import array2function
    fv = array2function(g,v)
    return Potential(fv) # return the potential





def commensurate_vortex_harmonic(g):
    """Return a skyrmion potential
    commensurate with the lattice.
    It may work only for triangular and square lattice
    """
    from . import geometry
    if g.dimensionality==2:
      a12 = g.a2.dot(g.a1)/(np.sqrt(g.a1.dot(g.a1))*np.sqrt(g.a1.dot(g.a1)))
      if 0.49<abs(a12)<0.51: 
          g0 = geometry.triangular_lattice() # triangular
          # this is a quick fix
      elif abs(a12)<0.01: 
          g0 = geometry.square_lattice() # square
      else: raise # not implemented
    else: raise
    h0 = g0.get_hamiltonian()
    h0.add_rashba(1.0) # add Rashba SOC
    r2f = g.get_k2K_generator() # get function
    hk = h0.get_hk_gen() # generator
    da = np.sqrt(g.a1.dot(g.a1))/2 # length of the vector
    def fun(r):
        rf = r/da
        rf = r2f(rf) # convert to fractional
#        rf = Ri@rf
        m = hk(rf) # evaluate Hamiltonian
        mx,my = m[0,1].real,-m[0,1].imag
        return np.array([mx,my,0.])
    return Potential(fun)


def commensurate_skyrmion_harmonic(g):
    """Create a skyrmion commensurate with the lattice"""
    # this function as nood as good as it could
    # probably the vectors should have the same length
    fv = commensurate_vortex(g) # get the vortex part
    vs = [fv(ri) for ri in g.r] # magnetizations
    vmax = np.max([vi.dot(vi) for vi in vs]) # absolute value
    fz = commensurate_potential(g,minmax=[-1.,1.],n=6,k=1) # z component
    def fun(r):
        m = fv(r) # xy component
        dz = vmax - m.dot(m) # value of z component
        return m + np.array([0.,0.,fz(r)])*dz
    return Potential(fun)



def commensurate_skyrmion(g,r0=None):
    """Return a conmmensurate skyrmion"""
    if r0 is not None: rc = r0
    else: rc = np.mean(g.r,axis=0) # average position
    da = np.sqrt(g.a1.dot(g.a1)) # average radius
    R = da/4. # radius of the skyrmion
    def fun(r):
        r = r - rc # radious with respect to the center
        r = r/R # normalize to the radious of the skyrmion
        dr = np.sqrt(r.dot(r)) # distance to center
        if dr<1.: 
            mz = np.sqrt(1. - dr**2) # z component
            return r + np.array([0.,0.,mz])
        if 1.<dr<1.5: 
            mxy = r/dr*(1.5-dr)
            mz = np.sqrt(1. - mxy.dot(mxy)) # z component
            return mxy + np.array([0.,0.,-mz])
        else:
            return np.array([0.,0.,-1.])
    return Potential(fun)



def circle(r0,v=0.,R=1.1,g=None,**kwargs):
    """Create the potential for a circle"""
    def f(r):
        dr = r-r0
        if dr.dot(dr)<R: return v
        else: return 0.
    out = Potential(f) # return the potential
    if g is None: return out
    else: return periodic_replicate(out,g,**kwargs)



def periodic_replicate(P,g,n=3):
    """Enforce a potential to have the periodicity
    in space of the lattice"""
    dl = g.neighbor_directions(n) # list with neighboring cells to take
    def fout(r):
        o = 0.
        for d in dl:
            ri = r + g.a1*d[0] + g.a2*d[1] + g.a3*d[2]
            o = o + P(ri) # add contribution
        return o
    return Potential(fout)







# Hamiltonians for second order topological insulators
import numpy as np

from .. import geometry 
def square_2OTI(delta=0.):
    """Return the Hamiltonian of a 2nd order TI in the square lattice"""
    d = delta # redefine
    g = geometry.square_lattice()
    g = g.get_supercell((2,2)) # make a supercell
    h = g.get_hamiltonian(has_spin=False)
    h.geometry.center_in_atom() # to get a real gauge
    m = h.intra*(1.-d) # decrease hopping
#    m[0,1] *= -1. # do it by hand
#    m[1,0] *= -1. # do it by hand
    h = h*(1+d) # scale all
    # this is should be written properly
    h.intra = m
    h.add_peierls(0.5)
    return h # return Hamiltonian



def chain_1OTI(delta=0.):
    """Return the Hamiltonian of a 2nd order TI in the square lattice"""
    d = delta # redefine
    g = geometry.chain()
    g = g.get_supercell(2) # make a supercell
    h = g.get_hamiltonian(has_spin=False)
    m = h.intra*(1.-d) # decrease hopping
    h = h*(1+d) # scale all
    # this is should be written properly
    h.intra = m
    return h # return Hamiltonian









import numpy as np

def discard_outliers(x,y):
    """Given x and y, discard outliers in y"""
    yn = y/np.median(y) # normalize to the median
    yp = np.percentile(y,90)
    xo = x[y<yp] # less than five times the median
    yo = y[y<yp] # less than five times the median
    return xo,yo





from scipy.linalg import eigvalsh
import numpy as np

def eigenvalues(h,nk):
    """Return all the eigenvalues of a Hamiltonian"""
    import klist
    h.turn_dense()
    ks = klist.kmesh(h.dimensionality,nk=nk) # get grid
    hkgen = h.get_hk_gen() # get generator
    e0 = 0.0
    import timing
    est = timing.Testimator(maxite=len(ks))
    for k in ks: # loop
      est.iterate()
      es = eigvalsh(hkgen(k)).tolist() # add
      e0 += np.sum(es[es<0.]) # add contribution
    return e0/len(ks) # return the ground state energy

#
#def gsenergy(h,nk=10):
#  """Calculate the ground state energy"""
#  if h.dimensionality!=2: raise  # continue if two dimensional
#  hk_gen = h.get_hk_gen() # gets the function to generate h(k)
#  kxs = np.linspace(0.,1.,nk)  # generate kx
#  kys = np.linspace(0.,1.,nk)  # generate ky
#  e0 = 0.0 # initalize
#  for x in kxs:
#    for y in kys:
#      hk = hk_gen([x,y]) # get hamiltonian
#      es = eigvalsh(hk) # eigenvalues
#      e0 += np.sum(es[es<0.]) # add contribution
#  return e0/nk**2



import numpy as np

def boundary_embedding_gf(EB,selfenergy=1.,energy=0.,delta=1e-4,**kwargs):
    """Given a Hamiltonian, return an Embedded object that has
    selfenergies at the boundaries of the system"""
    if EB.boundary_embedding_generator is None:
        EB.boundary_embedding_generator = selfe_generator(EB.H) # create 
    if callable(selfenergy): selfe = selfenergy(energy)
    else: selfe = selfenergy
    SE = selfe*EB.boundary_embedding_generator # selfenergy
    from .. import algebra
    iden = np.identity(SE.shape[0],dtype=np.complex) # identity
    emat = iden*(energy + delta*1j) # energy matrix
    gv = algebra.inv(emat - EB.H.intra - SE)   # Defective Green function
    return gv




def selfe_generator(H):
    """Return the matrix with the selfenergy"""
    from ..potentials import edge_potential
    fe = edge_potential(H.geometry) # get the edge potential
    HE = H*0. # make a copy
    HE.add_onsite(fe) # generate a dummy onsite energy
    SE = -1j*HE.intra # matrix acting as the selfenergy
    return SE # return the basic matrix




import numpy as np


def get_operator(self,name,**kwargs):
      """Return the conventional operator"""
      from . import operators
      from . import potentials
      if type(name) is operators.Operator: return name # return operator
      if type(name) is potentials.Potential or callable(name): 
          out = self.copy()*0. # initialize
          out.add_onsite(name) # add onsite
          return operators.Operator(out.intra,linear=True) # return operator
      if name is None: return None # return operator
      if name=="None": return None
      elif name in ["berry","Berry"]: 
          return operators.get_berry(self,**kwargs)
      elif name=="site": 
          return operators.get_site(self,**kwargs)
      elif name=="correlator": 
          return operators.get_correlator_ij(self,**kwargs)
      elif name=="valleyberry": 
          return operators.get_operator_berry(self,"valley",**kwargs)
      elif name=="szvalleyberry": 
          return operators.get_operator_berry(self,"sz_valley",**kwargs)
      elif name=="szberry": return operators.get_sz_berry(self,**kwargs)
      elif name in ["sx","Sx"]: return operators.get_sx(self)
      elif name in ["sy","Sy"]: return operators.get_sy(self)
      elif name in ["sz","Sz"]: return operators.get_sz(self)
      elif name=="location": # return a certain location
          return operators.get_location(self,**kwargs)
      elif name=="current": return operators.get_current(self)
      elif name=="energy": 
          return operators.Operator(self)
      elif name in ["bulk","Bulk"]: 
          return operators.get_bulk(self,**kwargs)
      elif name in ["edge","Edge","surface","Surface"]: 
          return operators.get_surface(self,**kwargs)
      elif name=="velocity": return operators.get_velocity(self)
      elif name=="sublattice": return operators.get_sublattice(self,mode="both")
      elif name=="sublatticeA": return operators.get_sublattice(self,mode="A")
      elif name=="sublatticeB": return operators.get_sublattice(self,mode="B")
      elif name=="interface": return operators.get_interface(self)
      elif name=="spair": return operators.get_pairing(self,ptype="s")
      elif name=="deltax": return operators.get_pairing(self,ptype="deltax")
      elif name=="deltay": return operators.get_pairing(self,ptype="deltay")
      elif name=="deltaz": return operators.get_pairing(self,ptype="deltaz")
      elif name=="electron": return operators.get_electron(self)
      elif name=="hole": return operators.get_hole(self)
      elif name=="mass": 
          from .mass import mass_operator
          return mass_operator(self,**kwargs)
      elif name in ["up"]: return operators.get_up(self)
      elif name in ["dn","down"]: return operators.get_dn(self)
      elif name=="zposition": return operators.get_zposition(self)
      elif name=="surface": return operators.get_surface(self)
      elif name=="ldos": 
          from . import ldos
          return ldos.ldos_projector(self,**kwargs)
      elif name=="yposition": return operators.get_yposition(self)
      elif name=="xposition": return operators.get_xposition(self)
      elif name=="electrons": return operators.get_electron(self)
      elif name=="mx":
        return self.get_operator("sx")@self.get_operator("electron")
      elif name=="my":
        return self.get_operator("sy")@self.get_operator("electron")
      elif name=="mz":
        return self.get_operator("sz")@self.get_operator("electron")
      elif name=="layer": return operators.get_layer(self,**kwargs)
      elif name=="valley":  return operators.get_valley(self,**kwargs)
      elif name=="valley_x":  return operators.get_valley_taux(self)
      elif name=="inplane_valley": return operators.get_inplane_valley(self)
      elif name=="valley_upper" or name=="valley_top":
        return operators.get_valley_layer(self,n=-1)
#      elif name=="inplane_valley_upper":
#        print("This operator only makes sense for TBG")
#        ht = self.copy()
#        ht.geometry.sublattice = self.geometry.sublattice * (np.sign(self.geometry.z)+1.0)/2.0
#        return operators.get_inplane_valley(ht)
      elif name=="valley_lower" or name=="valley_bottom":
        return operators.get_valley_layer(self,n=0)
      elif name in ["ipr","IPR"]: return operators.ipr
      elif name=="potential":  return operators.get_potential(self,**kwargs)
      elif name=="unfold":  
          from .unfolding import bloch_projector
          return bloch_projector(self,**kwargs)
      elif name=="singlet":  
          from .sctk.operator import real_singlet
          return real_singlet(self)
      elif self.has_kondo:
          from .specialhamiltoniantk import heavyfermion
          return heavyfermion.get_operator(self,name,**kwargs)
      else: raise


get_scalar_operator = get_operator




# rountines to create hamiltonians ased on a certain skeleton,
# used to build ribbons

from scipy.sparse import csc_matrix
from scipy.sparse import bmat
from . import sculpt
from . import geometry
from . import multicell
import numpy as np

def build_ribbon(hin,g=None,n=20):
  """ Build a supercell using a certain geometry as skeleton"""
  if g is None: # if skeleton not provided
    h.geometry.get_lattice_name()
    if h.geometry.lattice_name=="square": # square lattice
      g = geometry.square_ribbon(n) 
    else: raise # not implemented
  # now build the hamiltonian
  h = hin.copy() # generate hamiltonian
  # if the hamiltonian is not multicell, turn it so
  if not h.is_multicell: h = multicell.turn_multicell(h)
  gin = h.geometry # geometry of the hamiltonian input
# use the same axis
  gh = sculpt.rotate_a2b(h.geometry,h.geometry.a1,np.array([0.,1.,0.])) 
#  if np.abs(h.geometry.a1.dot(h.geometry.a2)) > 0.01: raise # orthogonal
#  gh = h.geometry
  def normalize(v): # normalize a vector
    return v/np.sqrt(v.dot(v))
# get reciprocal vectors
  (w1,w2,w3) = sculpt.reciprocal(normalize(gh.a1),normalize(gh.a2)) 
#  exit()
  def get_rij(r):
    """Provide a vector r, return this vector expressed in the basis cell""" 
    i = r.dot(w1) # first projection
    j = r.dot(w2) # second projection
    i,j = round(i),round(j)
    return [i,j,0] # return indexes
  ho = h.copy() # generate hamiltonian
  intra = [[None for i in range(len(g.r))] for j in range(len(g.r))]
  hoppings = [] # empty list for the hoppings
  for i in range(len(g.r)): # loop over positions
    intra[i][i] = h.intra # intracell hopping
  for i in range(len(g.r)): # hopping up to third cells
    for j in range(len(g.r)): # hopping up to third cells
      rij = g.r[i] - g.r[j] # distance between replicas
      intra[i][j] = multicell.get_tij(h,rij=get_rij(rij)) 
  ho.intra = csc_matrix(bmat(intra)) # add the intracell matrix
  for nn in [-3,-2,-1,1,2,3]: # hopping up to third cells
    inter = [[None for i in range(len(g.r))] for j in range(len(g.r))]
    for i in range(len(g.r)): # hopping up to third cells
      for j in range(len(g.r)): # hopping up to third cells
        rij = g.r[i] - g.r[j] # distance between replicas
        rij += nn*h.geometry.a1 # add the displacement
        if i==j: # for diagonal, at least zeros
          mm = multicell.get_tij(h,rij=get_rij(rij)) 
          if mm is None: inter[i][j] = h.intra*0.0 # store zero
          else: inter[i][j] = mm # store matrix
        else: inter[i][j] = multicell.get_tij(h,rij=get_rij(rij)) 
    hopping = multicell.Hopping() # create object
    hopping.m = csc_matrix(bmat(inter)) # store matrix
    hopping.dir = np.array([nn,0.,0.]) # store vector
    hoppings.append(hopping) # store hopping
  gout = g.copy() # copy geometry for the hamiltonian
  rs = []
  for jr in g.r: # loop over skeleton geometry
    for ir in h.geometry.r: # loop over basis
      rs.append(ir + jr[0]*h.geometry.a1 + jr[1]*h.geometry.a2)
  gout.r = np.array(rs) # store
  gout.r2xyz() # update
  gout.celldis = h.geometry.a1[0] # this has to be well done
  ho.geometry = gout # assign geometry
  ho.hopping = hoppings # store the full hoppings list
  ho.dimensionality = 1 # one dimensional
  ho.is_multicell = True # multicell Hamiltonian
  ho.is_sparse = True # sparse Hamiltonian
  return ho # return ribbon hamiltonian






def build_island(h,n=5,angle=30,nedges=6):
  """ Build an island starting from a 2d geometry"""
  gin = geometry.triangular_lattice() # create lattice
  angle = sculpt.get_angle(h.geometry.a1,h.geometry.a2)/np.pi*180
  if np.abs(angle-60)<1.: gin.a2 = -gin.a2 # change the unit cell
  g = sculpt.build_island(gin,n=n,angle=angle,nedges=nedges,clear=False) # get the island
  angle2 = sculpt.get_angle(gin.a1,gin.a2)/np.pi*180
  if np.abs(angle-angle2)>1.: raise # error in the angles
  gh = sculpt.rotate_a2b(h.geometry,h.geometry.a1,gin.a1) # use the same axis
  # now define a function to select the correct hopping
  (w1,w2,w3) = sculpt.reciprocal(gin.a1,gin.a2) # get reciprocal vectors
  def get_rij(r):
    """Provide a vector r, return this vector expressed in the basis cell""" 
    i = r.dot(w1) # first projection
    j = r.dot(w2) # second projection
    return i,j # return indexes
  # turn into multicell
  if not h.is_multicell: h = multicell.turn_multicell(h) 
  h.turn_sparse() # convert into sparse
  intra = [[None for ri in g.r] for rj in g.r ]
  # loop over the skeleton
  for i in range(len(g.r)): # loop over i
    ri = g.r[i]
    for j in range(len(g.r)): # loop over j
      rj = g.r[j]
      vi,vj = get_rij(ri-rj) # get the vector
      if np.abs(vi)>2 or np.abs(vj)>2: continue
      m = multicell.get_tij(h,rij=np.array([vi,vj,0])) # return the matrix
      intra[i][j] = m # store in the matrix
  # fic the new hamiltonian
  ho = h.copy() # copy hamiltonian object
  from scipy.sparse import bmat
  ho.dimensionality = 0 # zero dimensional
  ho.intra = bmat(intra) # store hamiltonian
  # fix the new geometry
  go = h.geometry.copy() # copy the original geometry
  go.dimensionality = 0 # zero dimensional
  rs = [] # empty list
  for rd in g.r:
    for r in h.geometry.r:
      vi,vj = get_rij(rd) # get the vector
      ri = r +vi*h.geometry.a1 + vj*h.geometry.a2
      rs.append(ri) # append the vector
  rs = np.array(rs) # convert to array
  go.r = rs # store
  if go.atoms_have_names: # if the atoms have names, expand
    go.atoms_names = go.atoms_names*len(g.r) # enlarge the list
  go.r2xyz() # fill the xyz values
  ho.geometry = go # store in the hamiltonian
  go.write()
  return ho




def image2island(impath,h,s=20):
  """ Build an island starting from a 2d geometry"""
  gin = h.geometry # create lattice
  g = sculpt.image2island(impath,gin,s=s) # get the island
  angle = 0.0
  angle2 = sculpt.get_angle(gin.a1,gin.a2)/np.pi*180
  if np.abs(angle-angle2)>1.: raise # error in the angles
  gh = sculpt.rotate_a2b(h.geometry,h.geometry.a1,gin.a1) # use the same axis
  # now define a function to select the correct hopping
  (w1,w2,w3) = sculpt.reciprocal(gin.a1,gin.a2) # get reciprocal vectors
  def get_rij(r):
    """Provide a vector r, return this vector expressed in the basis cell""" 
    i = r.dot(w1) # first projection
    j = r.dot(w2) # second projection
    return i,j # return indexes
  # turn into multicell
  if not h.is_multicell: h = multicell.turn_multicell(h) 
  h.turn_sparse() # convert into sparse
  intra = [[None for ri in g.r] for rj in g.r ]
  # loop over the skeleton
  for i in range(len(g.r)): # loop over i
    ri = g.r[i]
    for j in range(len(g.r)): # loop over j
      rj = g.r[j]
      vi,vj = get_rij(ri-rj) # get the vector
      if np.abs(vi)>2 or np.abs(vj)>2: continue
      m = multicell.get_tij(h,rij=np.array([vi,vj,0])) # return the matrix
      intra[i][j] = m # store in the matrix
  # fic the new hamiltonian
  ho = h.copy() # copy hamiltonian object
  from scipy.sparse import bmat
  ho.dimensionality = 0 # zero dimensional
  ho.intra = bmat(intra) # store hamiltonian
  # fix the new geometry
  go = h.geometry.copy() # copy the original geometry
  go.dimensionality = 0 # zero dimensional
  rs = [] # empty list
  for rd in g.r:
    for r in h.geometry.r:
      vi,vj = get_rij(rd) # get the vector
      ri = r +vi*h.geometry.a1 + vj*h.geometry.a2
      rs.append(ri) # append the vector
  rs = np.array(rs) # convert to array
  go.r = rs # store
  if go.atoms_have_names: # if the atoms have names, expand
    go.atoms_names = go.atoms_names*len(g.r) # enlarge the list
  go.r2xyz() # fill the xyz values
  ho.geometry = go # store in the hamiltonian
  go.write()
  return ho






import numpy as np

def canonical_unitary(self,fin):
    """Return a function that wraps any matrix inot ints canonical unitary"""
    def fun(k):
        frac_r = self.geometry.frac_r # fractional coordinates
        frac_r = frac_r - frac_r[0]
        fphase = lambda ri: np.exp(1j*np.pi*2.*ri.dot(k)) # phase
        U = np.diag([fphase(r) for r in frac_r])
        U = np.array(U) # this is without .H
        U = self.spinless2full(U) # increase the space if necessary
        hk = fin(k)
        Ud = np.conjugate(U.T) # dagger
        hk = Ud@hk@U
        return hk # return matrix
    return fun



def canonical_unitary_generator(self):
    """Return a function that wraps any matrix inot ints canonical unitary"""
    raise
    def fun(k=None,d=None):
        """Given a certain directional hopping, return the Bloch matrix"""
        frac_r = self.geometry.frac_r # fractional coordinates
        frac_r = frac_r - frac_r[0] + np.array(d) # new fractional coordinates
        fphase = lambda ri: np.exp(1j*np.pi*2.*ri.dot(k)) # phase
        U = np.diag([fphase(r) for r in frac_r])
        U = np.array(U) # this is without .H
        U = self.spinless2full(U) # increase the space if necessary
        return U
    return fun # return the generating function

def canonical_unitary_hopping(self):
    """Return the generator for the Bloch hoppings"""
    fb = canonical_unitary_generator(self) # generating unitaries
    def fun(m,d=[0,0,0],**kwargs):
        U1 = fb(d=[0,0,0],**kwargs) # get unitary
        U2 = fb(d=d,**kwargs) # get unitary
        U1d = np.conjugate(U1).T
        U2d = np.conjugate(U2).T
        return U1d@m@U2 # return unitary
    return fun



import numpy as np
import scipy.linalg as lg
from numba import jit


def matrix_tails(m,discard=None):
    """Return the tails of a 1d Hamiltonian"""
    es,ws = lg.eigh(m) # diagonalize matrix
    ws = np.abs(np.transpose(ws)) # wavefunctions
    if discard is not None: # if there is a discard function
      wout = []
      eout = []
      for (e,w) in zip(es,ws): # loop over waves
        if not discard(w): # check if this wave is discarded 
          wout.append(w) # store wave
          eout.append(e) # store wave
      ws = np.array(wout) # store
      es = np.array(eout) # store
    ls = np.array([loclength([w]) for w in ws]) # localization length
#    np.savetxt("TAILS.OUT",np.matrix([es,ls]).T)
    return (es,ls) # return data
    



def tails(vs):
  """Return the log of the tails, centered around the maximum"""
  return tails_python(vs) # python function



def loclength(vs):
    """Return the log of the average value of the tails"""
    dsv = tails(vs) # return all the tails
    def get(cutoff):
        """Compute the tails for a single cutoff"""
        ds = np.mean(dsv,axis=0) # average over waves
        # this is a wway of discardinig too smmall elements
        # there could be better ways of doing it
        ds = ds[ds>cutoff] # only sizable elements
        # now do the fit
        out = np.log(ds) # log of the density
        ns = 1.+np.array(range(len(out))) # length
        ps = np.polyfit(ns,out,1) # make a fit
        return -ps[0] # return the slope
    cutoffs = [1e-6,1e-7,1e-8,1e-9,1e-10] # several cutoff
    cutoffs = [1e-10] # one cutoff
    ps = np.array([get(c) for c in cutoffs])
#    print(np.sqrt(np.mean((np.mean(ps)-ps)**2)))
    return np.mean(ps) # return the average








def tails_python(vs):
    """Python implementation, Return the log of the tails, 
    centered around the maximum"""
    vs = np.array(vs)
    out = np.zeros((vs.shape[0],vs.shape[1]//2)) 
    return tails_python_jit(vs,out)



@jit(nopython=True)
def tails_python_jit(vs,out):
    """Python implementation, Return the log of the tails, 
    centered around the maximum"""
    ii = 0 # counter
    for v in vs: # loop over wavefunctions
        d = (v*np.conjugate(v)).real # density
        n = len(d) # size of the array
        nmax = np.argmax(v) # index with the maximum
        ds = np.zeros(n//2) # initialize logd
        for j in range(n//2): # loop over components
            ir = (nmax + j)%n # right index
            il = (nmax - j)%n # left index
            ds[j] = d[ir] + d[il] # average tails
        out[ii,:] = ds[:]
        ii += 1 # increase counter
    return out # return array





# library to calculate transpot in multiterminal devices

import numpy as np
from . import neighbor

class Device():
  """ Device with leads and scattering part"""
  def __init__(self):
    self.leads = [] # empty list of leads
  def biterminal(self,right_g=None,left_g=None,central_g=None,fun=None,
                    disorder=0.0):
    """Create the matrices for a biterminal device, based on geometries"""
    if fun is None: # no function provided
      def fun(r1,r2):
        dr = r1-r2
        if .7<dr.dot(dr)<1.3: return True
        else: return False
    leadr = Lead() # right lead
    leadl = Lead() # left lead
    Rr = right_g.r # positions
    Lr = left_g.r # positions
    Cr = central_g.r # positions
    leadr.intra = neighbor.parametric_hopping(Rr,Rr,fun) # intra term
    leadl.intra = neighbor.parametric_hopping(Lr,Lr,fun) # intra term
    intra = neighbor.parametric_hopping(Cr,Cr,fun) # intra term
    for i in range(intra.shape[0]): # add disorder
      intra[i,i] += disorder*(np.random.random()-.5)
    self.intra = intra # store
    leadr.coupling = neighbor.parametric_hopping(Rr,Cr,fun) # coupling
    leadl.coupling = neighbor.parametric_hopping(Lr,Cr,fun) # coupling
    # now coupling within the lead
    Rr_dis = [r-right_g.a1 for r in Rr] # displace
    Lr_dis = [r-left_g.a1 for r in Lr] # displace
    leadr.inter = neighbor.parametric_hopping(Rr,Rr_dis,fun) # intra term
    leadl.inter = neighbor.parametric_hopping(Lr,Lr_dis,fun) # intra term
    # store positions
    leadr.r = Rr
    leadl.r = Lr
    self.r = Cr
    # store leads
    self.leads = [leadr,leadl] # store the leads
  def write(self):
    """Write positions of the atoms"""
    np.savetxt("CENTRAL.XYZ",self.r) # write central
    for i in range(len(self.leads)): 
      np.savetxt("LEAD_"+str(i)+".XYZ",self.leads[i].r) # write lead
    
  def transmission(self,energy=0.0):
    """Calculate the transmission"""
    return landauer(self,energy)
  def write_current(self,energy=0.0):
    """Calculate the transmission"""
    den = central_density(self,energy=energy)
    from .ldos import write_ldos
    write_ldos(self.r[:,0],self.r[:,1],den,output_file="CURRENT.OUT")



class Lead():
  """ Class for a lead"""
  intra = None  # intraterm
  inter = None  # interterm
  coupling = None  # coupling to the center
  def get_green(self,energy,error=0.00001,delta=0.00001):
    """ Get surface green function"""
    from . import green 
    grb,gr = green.green_renormalization(self.intra,self.inter,error=error,
                                          energy=energy,delta=delta)
    return gr
  def get_selfenergy(self,energy,error=0.0001,delta=0.0001):
    """ Get selfenergy"""
    gr = self.get_green(energy,error=error,delta=delta) # get greenfunction
    t = self.coupling # coupling
    selfenergy = t.H * gr * t 
    return selfenergy


def landauer(d,energy,ij=[(0,1)],error=0.000001,delta=0.00001):
  """ Calculate landauer tranmission between leads i,j """
  Ms = landauer_matrix(d,energy,ij=ij,error=error,delta=delta) # matrix
  Ts = [m.trace()[0,0] for m in Ms]
  return Ts


def landauer_matrix(d,energy,ij=[(0,1)],error=0.000001,delta=0.00001):
  """Return the Landauer matrices"""
  ss = [l.get_selfenergy(energy,error=error,delta=delta) for l in d.leads]
  # sum of selfenergies
  ssum = ss[0]*0.  
  for s in ss:  ssum += s # sum the selfenergies
  # identity matrix
  iden = np.identity(d.intra.shape[0])
  # calculate central green function
  gc = ((energy + delta*1j)*iden -  d.intra - ssum). I
  # calculate transmission
  ts = [] # empty list
  for (i,j) in ij: # loop over pairs
    # calculate spectral functions of the leads
    gammai = 1j*(ss[i]-ss[i].H) # gamma function
    gammaj = 1j*(ss[j]-ss[j].H) # gamma function
    # calculate transmission
    t = (gammai*gc*gammaj.H*gc.H).real 
    ts.append(t) # add to the list
  return ts # return list




def central_density(d,energy,ij=[(0,1)],error=0.000001,delta=0.00001):
  """Return the Landauer matrices"""
  ss = [l.get_selfenergy(energy,error=error,delta=delta) for l in d.leads]
  # sum of selfenergies
  ssum = ss[0]*0.  
  for s in ss:  ssum += s # sum the selfenergies
  # identity matrix
  iden = np.identity(d.intra.shape[0])
  # calculate central green function
  gc = ((energy + delta*1j)*iden -  d.intra - ssum). I
  # calculate transmission
  den = np.array([gc[i,i].imag for i in range(gc.shape[0])])
  return den # return list

